# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

# AUTHOR: Parts of this file copied from PTK sources, by Jonathan Slenders.
#   https://github.com/prompt-toolkit/python-prompt-toolkit
#     examples/full-screen/text-editor.py
#   Wholly: class MessageDialog, and function show_dialog_as_float.
#   Partially: function show_message.
#
# LICENSE:
#   Copyright (c) 2014, Jonathan Slenders
#   All rights reserved.
#
#   Redistribution and use in source and binary forms, with or without modification,
#   are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#   * Redistributions in binary form must reproduce the above copyright notice, this
#     list of conditions and the following disclaimer in the documentation and/or
#     other materials provided with the distribution.
#
#   * Neither the name of the {organization} nor the names of its
#     contributors may be used to endorse or promote products derived from
#     this software without specific prior written permission.
#
#   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
#   ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#   ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

""""""

from gettext import gettext as _

from asyncio import ensure_future, Future

from prompt_toolkit.application.current import get_app
from prompt_toolkit.layout.containers import Float, HSplit
from prompt_toolkit.layout.dimension import D
from prompt_toolkit.widgets import Button, Dialog, Label

__all__ = (
    'alert_and_question',
    'show_message',
    # Private:
    #   'AlertResponseDialog',
    #   'MessageDialog',
    #   'show_dialog_as_float',
)


def show_message(root_container, title, text):
    async def coroutine():
        dialog = MessageDialog(title, text)
        await ensure_future(show_dialog_as_float(root_container, dialog))

    ensure_future(coroutine())


# ***

class MessageDialog(object):
    def __init__(self, title, text):
        self.future = Future()

        def set_done():
            self.future.set_result(None)

        ok_button = Button(text=_('OK'), handler=(lambda: set_done()))

        self.dialog = Dialog(
            title=title,
            body=HSplit([
                Label(text=text),
            ]),
            buttons=[ok_button],
            width=D(preferred=80),
            modal=True,
        )

    def __pt_container__(self):
        return self.dialog


async def show_dialog_as_float(root_container, dialog):
    " Coroutine. "
    float_ = Float(content=dialog)
    root_container.floats.insert(0, float_)

    app = get_app()

    focused_before = app.layout.current_window
    app.layout.focus(dialog)
    result = await dialog.future
    app.layout.focus(focused_before)

    if float_ in root_container.floats:
        root_container.floats.remove(float_)

    return result


# ***

class AlertResponseDialog(object):
    def __init__(self, title='', label_text='', prompt_ok='', prompt_no=''):
        # MAGIC_NUMBER: 4: Pad button label so appears padded ``< LIKE SO >``
        BUTTON_PADDING = 4

        self.future = Future()

        def accept_text(buffer):
            get_app().layout.focus(button_ok)
            self.text_area.buffer.complete_state = None

        def on_button_ok():
            self.future.set_result(True)

        def on_button_no():
            self.future.set_result(False)

        button_ok = Button(
            text=prompt_ok,
            handler=on_button_ok,
            width=(len(prompt_ok) + BUTTON_PADDING),
        )
        button_no = Button(
            text=prompt_no,
            handler=on_button_no,
            width=(len(prompt_no) + BUTTON_PADDING),
        )

        self.dialog = Dialog(
            title=title,
            body=HSplit([
                Label(text=label_text),
            ]),
            buttons=[button_ok, button_no],
            width=D(preferred=80),
            modal=True,
        )

    def __pt_container__(self):
        return self.dialog


def alert_and_question(
    root_container,
    title='',
    label_text='',
    prompt_ok=_('OK'),
    prompt_no=_('Cancel'),
    on_close=lambda x: None,
):
    async def coroutine():
        ar_dialog = AlertResponseDialog(
            title=title,
            label_text=label_text,
            prompt_ok=prompt_ok,
            prompt_no=prompt_no,
        )
        result = await show_dialog_as_float(root_container, ar_dialog)
        on_close(result)

    ensure_future(coroutine())




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Manages User file containing regex for filtering completions/suggestions."""

import os
import re

from gettext import gettext as _

from dob_bright.termio import dob_in_user_warning

from dob_prompt.prompters.enc_actegory_re import RegExpActegory

__all__ = (
    'load_no_completion',
    'NoComplete',
)


class NoComplete(object):
    """"""

    def __init__(self, re_act=None, re_cat=None, re_tag=None):
        self.re_act = re_act
        self.re_cat = re_cat
        self.re_tag = re_tag


def load_no_completion(controller):
    """"""

    class SectionState(object):
        """File parsing state machine."""

        section_activity = re.compile(r'^\s*\[activity\]\s*$')
        section_category = re.compile(r'^\s*\[category\]\s*$')
        section_tagnames = re.compile(r'^\s*\[tag\]\s*$')

        def __init__(self):
            self.state = None

        def changes_state(self, rule):
            for possible_section in (
                SectionState.section_activity,
                SectionState.section_category,
                SectionState.section_tagnames,
            ):
                # We can use match() because only care about start of string.
                if possible_section.match(rule):
                    self.state = possible_section
                    return True
            return False

        @property
        def on_activity(self):
            return self.state is SectionState.section_activity

        @property
        def on_category(self):
            return self.state is SectionState.section_category

        @property
        def on_tagnames(self):
            return self.state is SectionState.section_tagnames

    # ***

    affirm = controller.affirm
    config = controller.config

    _activities = []
    _categories = []
    _tagsofname = []

    comment_line = re.compile(r'^\s*#')
    comment_decode = re.compile(r'^(\s*)\\#')

    nothing_but_blankness = re.compile(r'^\s*$')

    matches_nothing = re.compile(r'[^\w\W]')
    nothing_matches = 3 * [matches_nothing]

    sep_sym = '@'
    re_actegory = RegExpActegory(sep=sep_sym)

    state = SectionState()

    def decode_leading_comment(text):
        return comment_decode.sub('\\1#', text)

    def _load_no_completion():
        try:
            re_act, re_cat, re_tag = re_compiled_user_ignores()
        except Exception as err:
            # FIXME/2019-11-30: (lb): Better error message.
            dob_in_user_warning(str(err))
            re_act, re_cat, re_tag = nothing_matches
        return NoComplete(re_act, re_cat, re_tag)

    def re_compiled_user_ignores():
        ignore_fpath = ignores_file_path()
        compiled_re = ignores_file_parse(ignore_fpath)
        return compiled_re

    def ignores_file_path():
        ignore_fpath = config['editor.ignore_fpath']
        if not os.path.exists(ignore_fpath):
            return None
        return ignore_fpath

    def ignores_file_parse(ignore_fpath):
        if not ignore_fpath:
            return nothing_matches
        return open_and_parse(ignore_fpath)

    def open_and_parse(ignore_fpath):
        open_and_consume_rules(ignore_fpath)
        return process_consumed_rules()

    def open_and_consume_rules(ignore_fpath):
        with open(ignore_fpath, 'r', encoding='utf-8') as ignore_fpath_f:
            for line in ignore_fpath_f:
                consume_if_not_empty_line(line.rstrip('\n'))

    def consume_if_not_empty_line(rule):
        if nothing_but_blankness.match(rule):
            return
        consume_if_not_comment(rule)

    def consume_if_not_comment(rule):
        if comment_line.match(rule):
            return
        cleaned_rule = decode_leading_comment(rule)
        check_state_maybe_consume(cleaned_rule)

    def check_state_maybe_consume(rule):
        if state.changes_state(rule):
            return
        consume_completion_exclusion_rule(rule)

    def consume_completion_exclusion_rule(rule):
        if state.on_tagnames:
            _tagsofname.append(rule)
        else:
            activity, category = re_actegory.split_parts(rule)
            consume_actegory_rule(activity, category, rule)

    def consume_actegory_rule(activity, category, rule):
        if not activity and not category:
            # User has a solo '@' in their ignore file. Ignore it.
            return
        if state.on_activity:
            if not activity:
                # User has '@foo' in [activity] section. Interpret as category.
                _categories.append(category)
            else:
                _activities.append(rule)
        elif state.on_category:
            if not category:
                # There was no @ in the name, so was category anyway.
                category = activity
            if category:
                _categories.append(category)
            else:
                affirm(False)
        else:
            # Haven't seen first section yet!
            # FIXME/2019-11-30: (lb): Better error message.
            dob_in_user_warning(_(
                'Cannot discern no-completion rule seen before first section: ‘{}’'
            ).format(rule))

    def process_consumed_rules():
        re_act = compile_rules(_activities)
        re_cat = compile_rules(_categories)
        re_tag = compile_rules(_tagsofname)
        return re_act, re_cat, re_tag

    def compile_rules(rules):
        if not rules:
            return matches_nothing
        return re.compile(r'^({})$'.format('|'.join(rules)))

    # ***

    return _load_no_completion()




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Facts Carousel"""

from prompt_toolkit.formatted_text import merge_formatted_text
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.keys import Keys
from prompt_toolkit.shortcuts import PromptSession

__all__ = (
    'confirm',
    # Private:
    #  'create_confirm_session',
)

# (lb): This is ripped from PPT's shortcuts/prompt.py


def create_confirm_session(message, suffix=' (y/n) ', **kwargs):
    """
    Create a `PromptSession` object for the 'confirm' function.
    """
    assert isinstance(message, str)
    bindings = KeyBindings()

    count = {'cc': 0, }

    @bindings.add('y')
    @bindings.add('Y')
    def yes(event):
        count['cc'] = 0
        session.default_buffer.text = 'y'
        event.app.exit(result=True)

    @bindings.add('n')
    @bindings.add('N')
    # (lb): PPT's confirm behavior is to assume Ctrl-c means No,
    #   in which case you end up with the following flow in dob:
    #       Ctrl-c => Break out of Application and Bring up confirmation;
    #       Ctrl-c => Confirmation returns No, so re-run Application.
    #   2018-06-17: I'm not sold on any particular work-flow, but I think
    #   I'd like Ctrl-c -- if you mash it enough -- to finally break out.
    #   Or, at least. I think Ctrl-c followed by Ctrl-c should not act as
    #   a toggle; the second (and subsequent) Ctrl-c's should be ignored,
    #   if nothing else.
    # @bindings.add('c-c')
    def no(event):
        count['cc'] = 0
        session.default_buffer.text = 'n'
        event.app.exit(result=False)

    # (lb): 2018-06-17: Here's the workflow I'd think.
    #   First Ctrl-c: Break out of Application, and bring up quit-okay? prompt.
    #   Second Ctrl-c: Ignored (increment count from 0 to 1).
    #   Third Ctrl-c: Bingo! Really exit.
    # MAYBE: Keep temp file of edits and tell user where/how they can recover.
    @bindings.add('c-q')
    def mash(event):
        count['cc'] += 1
        if count['cc'] > 1:
            event.app.exit(result=True)

    @bindings.add(Keys.Any)
    def _(event):
        """Disallow inserting other text."""
        pass

    complete_message = merge_formatted_text([message, suffix])
    session = PromptSession(complete_message, key_bindings=bindings, **kwargs)
    return session


def confirm(message='Confirm?', suffix=' (y/n) ', **kwargs):
    """
    Display a confirmation prompt that returns True/False.
    """
    session = create_confirm_session(message, suffix, **kwargs)
    return session.prompt()




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Facts Carousel Status line (footer)"""

import time

from gettext import gettext as _

from prompt_toolkit.layout.containers import to_container
from prompt_toolkit.widgets import Label

__all__ = (
    'ZoneLowdown',
)


class ZoneLowdown(object):
    """"""
    def __init__(self, carousel):
        self.carousel = carousel
        self.hot_notif = ''
        self.notif_expiry = None

    def standup(self):
        self.status_label = Label(text='', style='class:footer')
        self.label_container = to_container(self.status_label)

    def update_status(self, hot_notif, clear_after_secs=None):
        self.hot_notif = hot_notif
        self.status_label.text = hot_notif
        self.reset_notif_expiry(clear_after_secs)

    LOWDOWN_NOTIFY_MESSAGE_LIFETIME_SECS = 2.71828

    def reset_notif_expiry(self, clear_after_secs=None):
        if clear_after_secs is None:
            clear_after_secs = ZoneLowdown.LOWDOWN_NOTIFY_MESSAGE_LIFETIME_SECS
        if not clear_after_secs:
            self.notif_expiry = None
        else:
            self.notif_expiry = time.time() + clear_after_secs

    # ***

    def reset_status(self):
        self.hot_notif = ''  # So format_lowdown_text prints PK.
        formatted_text = self.format_lowdown_text()
        self.notif_expiry = None
        self.update_status(hot_notif=formatted_text, clear_after_secs=0)

    def selectively_refresh(self):
        # Clear status messages after timeout.
        if self.notif_expiry and time.time() >= self.notif_expiry:
            self.reset_status()

    # ***

    def rebuild_viewable(self):
        """"""
        self.status_label.text = self.format_lowdown_text()
        return self.label_container

    def format_lowdown_text(self):
        """"""
        def _format_lowdown_text():
            # If message notif, show that, until next reset_showing_help,
            # else show current Fact ID and binding hints.
            return show_fact_id_and_binding_hints()

        def show_fact_id_and_binding_hints():
            showing_text = self.hot_notif or showing_fact()
            helpful_text = "[?]: Help / [C-S]: Save / [C-Q]: Quit"

            pad_len = (
                self.carousel.avail_width
                - len(showing_text)
                - len(helpful_text)
            )
            padding = ' ' * pad_len

            hot_notif_or_fact_id_style = self.carousel.add_stylable_classes(
                ppt_widget=None, friendly_name='footer-fact-id',
            )
            # MEH/2019-12-02: (lb): I don't feel like styling the whole bottom line...

            formatted = [
                (hot_notif_or_fact_id_style, showing_text),
                ('', padding),
                ('', helpful_text),
            ]

            return formatted

        def showing_fact():
            curr_edit = self.carousel.edits_manager.curr_edit
            if 'interval-gap' in curr_edit.dirty_reasons:
                context = _('Gap')
                location = _("of {0}").format(curr_edit.format_delta(style=''))
                deleted = _(' [edit to add]')
            else:
                if curr_edit.pk > 0:
                    # 2019-01-26: (lb): For parallelism, I had a similar prefix,
                    #   context = _('Old')
                    # here, but I think it looks funny to call a Fact "Old".
                    context = ''
                    location = _("ID #{0}").format(curr_edit.pk)
                else:
                    num_unstored = self.carousel.edits_manager.edit_fact_count
                    context = _('New')
                    location = _("{0:>{1}} of {2}").format(
                        self.carousel.edits_manager.edit_fact_index + 1,
                        len(str(num_unstored)),
                        num_unstored,
                    )
                deleted = _(' [del]') if curr_edit.deleted else ""
            text = _("{0}{1}Fact {2}{3}").format(
                context, ' ' if context else '', location, deleted,
            )
            return text

        return _format_lowdown_text()




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Zone Details End Time Code (for easier diff-meld against Start Time Code)"""

from datetime import timedelta

from .exceptions import catch_action_exception

__all__ = (
    'ZoneDetails_TimeEnd',
)


class ZoneDetails_TimeEnd(object):
    """"""

    def __init__(self):
        super(ZoneDetails_TimeEnd, self).__init__()

    def add_header_end_time(self):
        self.widgets_end = self.add_header_section(
            'end', 'end_fmt_local_nowwed', editable=True,
            mouse_handler=self.header_time_end_mouse_handler,
        )

    def header_time_end_mouse_handler(self, mouse_event):
        self.zone_manager.toggle_focus_time_end(event=None)
        self.zone_manager.ppt_control_buffer_avoid_selection(
            self.widgets_end.text_area.control,
            mouse_event,
        )

    def refresh_time_end(self):
        self.refresh_val_widgets(self.widgets_end)

    # ***

    @catch_action_exception
    def edit_time_end(self, event=None, focus=True):
        if focus:
            edit_fact = self.zone_manager.facts_diff.edit_fact
            # Use or-now, unlike start time, because there's on one possible
            # open-ended fact (without an end time) and it'd be the latest fact.
            end_fmt_local_or_now = edit_fact.end_fmt_local_or_now
            self.widgets_end.orig_val = end_fmt_local_or_now
            self.widgets_end.text_area.text = end_fmt_local_or_now
            self.edit_time_focus(self.widgets_end)
            return True
        else:
            return self.edit_time_leave(self.widgets_end)

    # ***

    def apply_edit_time_removed_end(self, edit_fact):
        if not edit_fact.end:
            # Already cleared; nothing changed. Return True to not show
            # dialog message; apply_edit_time_end is called but returns
            # immediately.
            return True
        if self.carousel.edits_manager.conjoined.is_final_fact(edit_fact):
            # First if-branch should've handled, not us!
            self.carousel.controller.affirm(False)
            return True

        edits_manager = self.carousel.edits_manager
        edit_next = edits_manager.editable_fact_next(edit_fact)
        if (
            'interval-gap' in edit_next.dirty_reasons
            and self.carousel.edits_manager.conjoined.is_final_fact(edit_next)
        ):
            # Use case: User sets end time of active Fact, then tabs away
            # and tabs back, presses Ctrl-U to clear the time entry, then
            # tabs away. Remove Gap Fact and make Fact active again.
            # - Return True so apply_edit_time_end is called.
            return True
        else:
            self.widgets_end.text_area.text = edit_fact.end_fmt_local_or_now
            return False

    # ***

    def apply_edit_time_end(self, edit_fact, edit_time, verify_fact_times):
        if edit_fact.end == edit_time:
            # Nothing changed.
            return False
        edited_facts = [edit_fact]
        edits_manager = self.carousel.edits_manager
        # Make undoable.
        was_fact = edit_fact.copy()
        undoable_facts = [was_fact]
        # After apply_edit_time_removed_end, for editable_fact_next to not die.
        was_time = edit_fact.end
        if edit_time is None:
            edit_fact.end = None
        # Prohibit completely shadowing other facts' time windows, but allow
        # changing one fact's times to shorten the times of prev or next fact.
        edit_next = edits_manager.editable_fact_next(edit_fact)
        best_time = edit_time
        if best_time and edit_next and edit_next.end and (best_time > edit_next.end):
            # Adjust adjacent fact's time width to be no less that fact_min_delta.
            min_delta = int(self.carousel.controller.config['time.fact_min_delta'])
            # We do not want to make momentaneous facts, so use at least 1 minute.
            min_delta = max(1, min_delta)
            best_time = edit_next.end - timedelta(minutes=min_delta)
            if edit_next and edit_next.start and (best_time < edit_next.start):
                best_time = edit_next.start
        edit_fact.end = best_time
        # Verify edit_fact.start < edit_fact.end.
        if not verify_fact_times(edit_fact):
            edit_fact.end = was_time
            return False
        # If the edited time encroached on the neighbor, or if the neighbor
        # is an unedited gap fact, edit thy neighbor.
        if edit_next:
            if ((
                edit_next.start
                and edit_fact.end
                and (edit_fact.end > edit_next.start)
            ) or ('interval-gap' in edit_next.dirty_reasons)
            ):
                undoable_facts.append(edit_next.copy())
                edit_next.start = was_time
                edited_facts.append(edit_next)
            if edit_fact.end == edit_next.start:
                edit_fact.next_fact = edit_next
                edit_next.prev_fact = edit_fact
            else:
                edit_fact.next_fact = None
                edit_next.prev_fact = None
        else:
            edit_fact.next_fact = None
        edits_manager.add_undoable(undoable_facts, what='header-edit')
        edits_manager.apply_edits(*edited_facts)
        if edit_time is None:
            # (lb): This feels a little kludgy, or like this belongs elsewhere...
            self.carousel.edits_manager.conjoined.pop_final_gap_fact()
        return True




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma,  2015-2016 Eric Goller.  All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import tempfile

import editor

from nark.items.activity import Activity

# Lazy-load AwesomePrompt to save ~0.1 seconds when not needed.
from dob_prompt import prompters

__all__ = (
    'ask_user_for_edits',
    'ask_edit_with_editor',
)


# ***

def ask_user_for_edits(
    controller,
    fact,
    always_ask=False,
    prompt_agent=None,
    restrict_edit='',
    no_completion=None,
):
    """
    """

    def _ask_user_for_edits():
        verify_always_ask()

        prompter = get_prompter()

        ask_act_cat(prompter, fact)

        ask_for_tags(prompter, fact)

        fact_ask_description(fact)

        return prompter

    # ***

    def verify_always_ask():
        assert always_ask in [
            True, False, 'actegory', 'tags', 'description',
        ]

    # ***

    def get_prompter():
        if prompt_agent is None:
            return prompters.path.AwesomePrompt(controller)
        else:
            assert isinstance(prompt_agent, prompters.path.PrompterCommon)
            return prompt_agent

    # ***

    def ask_act_cat(prompter, fact):
        filter_activity, filter_category = prepare_ask_act_cat(fact)
        if (
            (filter_activity and filter_category and always_ask is False)
            or ('' != restrict_edit and 'actegory' != restrict_edit)
        ):
            return

        no_completion_act = None
        no_completion_cat = None
        if no_completion is not None:
            no_completion_act = no_completion.re_act
            no_completion_cat = no_completion.re_cat

        act_name, cat_name = prompter.ask_act_cat(
            filter_activity,
            filter_category,
            no_completion_act=no_completion_act,
            no_completion_cat=no_completion_cat,
        )
        set_actegory(fact, act_name, cat_name)

    def prepare_ask_act_cat(fact):
        filter_activity = ''
        if fact.activity and fact.activity.name:
            filter_activity = fact.activity.name

        filter_category = ''
        if fact.activity and fact.activity.category and fact.activity.category.name:
            filter_category = fact.activity.category.name

        return filter_activity, filter_category

    def set_actegory(fact, act_name, cat_name):
        fact.activity = Activity.create_from_composite(act_name, cat_name)
        try:
            fact.activity = controller.activities.get_by_composite(
                fact.activity.name, fact.activity.category, raw=False,
            )
        except KeyError:
            pass

    # ***

    def ask_for_tags(prompter, fact):
        if (
            (fact.tags and always_ask is False)
            or ('' != restrict_edit and 'tags' != restrict_edit)
        ):
            return

        no_complete_tag = None
        if no_completion is not None:
            no_complete_tag = no_completion.re_tag

        chosen_tags = prompter.ask_for_tags(
            already_selected=fact.tags,
            activity=fact.activity,
            no_completion=no_complete_tag,
        )
        fact.tags_replace(chosen_tags)

    # ***

    def fact_ask_description(fact):
        if (
            (fact.description and always_ask is False)
            or ('' != restrict_edit and 'description' != restrict_edit)
        ):
            return

        # (lb): Strip whitespace from the description. This is how `git` works.
        # Not that we have to be like git. But it makes parsed results match
        # the input, i.e., it we didn't strip() and then re-parsed the non-
        # stripped description, the parser would strip, and we'd see a difference
        # between the pre-parsed and post-parsed description, albeit only
        # leading and/or trailing whitespace. (If we wanted to preserve whitespace,
        # we'd have to make the parser a little more intelligent, but currently
        # the parser strip()s while it parses, to simplify the parsing algorithm.)
        raw_description = ask_edit_with_editor(controller, fact, fact.description)
        fact.description = raw_description.strip()

    # ***

    return _ask_user_for_edits()


# ***

def ask_edit_with_editor(controller, fact=None, content=''):
    def _ask_edit_with_editor():
        contents = prepare_contents(content)
        filename = temp_filename()
        edited = run_editor(filename, contents)
        return edited

    def prepare_contents(content):
        content = content if content else ''
        # # FIXME: py2 compatible? Or need to six.b()?
        # #contents = six.b(str(content))  # NOPE: Has problems with Unicode, like: ½
        # contents = text_type(content).encode()
        # FIXME/2020-01-26: (lb): Verify no longer an issue.
        contents = str(content).encode()
        return contents

    def temp_filename():
        tmpfile = tempfile.NamedTemporaryFile(
            prefix=prepare_prefix(),
            suffix=prepare_suffix(),
        )
        filename = tmpfile.name
        return filename

    def prepare_prefix():
        # Vim names the terminal with the file's basename, which is
        # normally meaningless, e.g., "tmprvapy77w.rst (/tmp)", but
        # we can give the randomly-named temp file a prefix to make
        # the title more meaningful.
        prefix = None
        if fact is not None:
            # E.g., "17:33 07 Apr 2018 "
            timefmt = '%H:%M %d %b %Y '
            if fact.start:
                prefix = fact.start.strftime(timefmt)
            elif fact.end:
                prefix = fact.end.strftime(timefmt)
        return prefix

    def prepare_suffix():
        # User can set a suffix, which can be useful so, e.g., Vim
        # sees the extension and set filetype appropriately.
        # (lb): I like my Hamster logs to look like reST documents!
        suffix = controller.config['term.editor_suffix'] or None
        return suffix

    def run_editor(filename, contents):
        # NOTE: You'll find EDITOR features in multiple libraries.
        #       The UX should be indistinguishable to the user.
        #       E.g., we could use click's `edit` instead of editor's:
        #
        #           click.edit(text=None,
        #                      editor=None,
        #                      env=None,
        #                      require_save=True,
        #                      extension='.txt',
        #                      filename=None)
        #
        result = editor.edit(filename=filename, contents=contents)
        # FIXME/2018-05-10: (lb): Is this py2 compatible? (2018-06-29: Do I care?)
        edited = result.decode()
        # Necessary?:
        #   edited = result.decode('utf-8')
        return edited

    return _ask_edit_with_editor()




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

""""""

from gettext import gettext as _

from dob_bright.termio import attr, bg, fg

__all__ = (
    'conflict_prefix',
    'prepare_log_msg',
)


# ***

def conflict_prefix(prefix):
    return (
        '{}{}{}'
        .format(
            bg('medium_violet_red'),
            prefix,
            attr('reset'),
        )
    )


# ***

def prepare_log_msg(fact_or_dict, msg_content):
    def _prepare_log_msg():
        try:
            line_num = fact_or_dict['line_num']
            line_raw = fact_or_dict['line_raw']
        except KeyError:
            line_num = fact_or_dict['parsed_source.line_num']
            line_raw = fact_or_dict['parsed_source.line_raw']
        except TypeError:
            line_num = fact_or_dict.parsed_source.line_num
            line_raw = fact_or_dict.parsed_source.line_raw
        line_num = line_num or 0
        line_raw = line_raw or ''
        # NOTE: Using colors overrides logger's coloring, which is great!
        return _(
            '{}{}{}: {}{}: {}{} / {}{}{}\n\n{}: {}“{}”{}\n\n{}: {}{}{}'
            .format(
                attr('bold'),
                conflict_prefix(_('Problem')),
                attr('reset'),

                fg('dodger_blue_1'),
                _('On line'),
                line_num,
                attr('reset'),

                attr('underlined'),
                msg_content,
                attr('reset'),

                conflict_prefix(_('  Typed')),
                fg('hot_pink_2'),
                line_raw.strip(),
                attr('reset'),

                conflict_prefix(_(' Parsed')),
                fg('grey_78'),
                fact_or_dict,
                attr('reset'),
            )
        )

    return _prepare_log_msg()




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

from ..config.styling.classes_style import (
    load_classes_style,
    load_matches_style
)
from ..config.styling.content_lexer import load_content_lexer
from ..config.styling.no_completion import load_no_completion


__all__ = (
    'prompt_and_save_confirmer',
)


# ***

def prompt_and_save_confirmer(
    controller,
    edit_facts=None,
    orig_facts=None,
    backup_callback=None,
    dry=False,
    **kwargs,
):
    """"""

    classes_style = load_classes_style(controller)
    matches_style = load_matches_style(controller)
    content_lexer = load_content_lexer(controller)
    no_completion = load_no_completion(controller)

    # Lazy-load the carousel and save ~0.065s.
    from dob_viewer.traverser.carousel import Carousel

    carousel = Carousel(
        controller,
        edit_facts=edit_facts,
        orig_facts=orig_facts,
        dirty_callback=backup_callback,
        dry=dry,
        classes_style=classes_style,
        matches_style=matches_style,
        content_lexer=content_lexer,
        no_completion=no_completion,
    )

    ready_facts = carousel.gallop(**kwargs)

    # The Carousel forces the user to save to exit! So ready_facts is empty!!
    # Or the user quit without saving, which would also mean no ready facts.
    # - 2020-01-28: (lb): If I recall correctly, I had originally
    #   plumbed Carousel to have you *review* Facts being imported
    #   (and only those Facts; user did not see other Facts); but now
    #   user sees Carousel of all Facts, with imported Facts inserted,
    #   but not saved, and user saves through Carousel.
    #   - Which is fine, probably desirable. But because user might
    #     have insisted `not use_carousel`, this module still has
    #     code to save Facts independent of the Carousel.
    #     Nonetheless, the Carousel is still coded to return a Facts
    #     array...
    #     MAYBE/2020-01-28: Could probably remove gallop() return list,
    #                  I think nowadays it'll always be the empty list??
    controller.affirm(not ready_facts)

    # CLOSED_LOOP: (lb): whoa_nellie is (disabled) kludge to avoid abandoned
    # event loop RuntimeError (better solution is needed; but sleeping here
    # before app exit can anecdotally avoid issue. And by "anecdotal" I mean
    # not necessarily guaranteed, so not a proper solution (and not even a
    # proper "kludge" by more stringent standards)).
    # TESTME/2020-02-01: Is this still necessary? Especially in PTK3?
    carousel.whoa_nellie()

    # ***



# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""User configurable interactive editor styling settings loaders."""

from gettext import gettext as _

import os

from configobj import ConfigObj, ConfigObjError

from dob_bright.termio import dob_in_user_warning

from ...traverser import various_styles
from . import load_obj_from_internal

__all__ = (
    'load_classes_style',
    'load_matches_style',
    # PRIVATE:
    # 'create_configobj',
)


def load_classes_style(controller):
    config = controller.config

    # (lb): It's times like these -- adding a dict to get around scoping
    # when sharing a variable -- that I think a method (load_classes_style)
    # should be a class. But this works for now.
    load_failed = {'styles': False}

    def _load_classes_style():
        named_style = resolve_named_style()
        classes_dict = try_load_dict_from_user_styling(named_style)
        classes_style = instantiate_or_try_internal_style(named_style, classes_dict)
        return classes_style

    def resolve_named_style():
        cfg_key_style = 'editor.styling'
        return config[cfg_key_style]

    def try_load_dict_from_user_styling(named_style):
        styles_path = resolve_path_styles()
        if not os.path.exists(styles_path):
            return None
        return load_dict_from_user_styling(styles_path, named_style)

    def resolve_path_styles():
        cfg_key_fpath = 'editor.styles_fpath'
        return config[cfg_key_fpath]

    def load_dict_from_user_styling(styles_path, named_style):
        config_obj = create_configobj(styles_path, nickname='styles')
        if config_obj is None:
            load_failed['styles'] = True
        elif named_style in config_obj:
            classes_dict = config_obj[named_style]
            return classes_dict
        return None

    def instantiate_or_try_internal_style(named_style, classes_dict):
        if classes_dict is not None:
            controller.affirm(isinstance(classes_dict, dict))
            defaults = prepare_base_style(classes_dict)
            update_base_style(named_style, classes_dict, defaults)
            return defaults
        return load_internal_style(named_style)

    def prepare_base_style(classes_dict):
        # Load base-style (e.g., various_styles.default) to ensure
        # all keys present (and defaulted), and then update that.
        base_style = 'default'
        if 'base-style' in classes_dict:
            base_style = classes_dict['base-style'] or 'default'
        try:
            defaults = getattr(various_styles, base_style)()
        except AttributeError as err:  # noqa: F841
            # Unexpected, because of choices= on base-style @setting def.
            controller.affirm(False)
            defaults = various_styles.default()
        return defaults

    def update_base_style(named_style, classes_dict, defaults):
        try:
            defaults.update_gross(classes_dict)
        except Exception as err:
            msg = _("Failed to load style named “{0}”: {1}").format(
                named_style, str(err),
            )
            dob_in_user_warning(msg)

    def load_internal_style(named_style):
        # HARDCODED/DEFAULT: classes_style default: 'default' (Ha!).
        # - This style uses no colors, so the UX will default to however
        #   the terminal already looks.
        default_style = 'default'
        classes_style_fn = load_obj_from_internal(
            controller,
            obj_name=named_style,
            internal=various_styles,
            default_name=default_style,
            warn_tell_not_found=not load_failed['styles'],
        )
        # If None, Carousel will eventually set to a default of its choosing.
        # - (lb): Except that we specified default_name, so never None:
        controller.affirm(classes_style_fn is not None)
        return classes_style_fn and classes_style_fn() or None

    # ***

    return _load_classes_style()


# ***

def load_matches_style(controller):
    config = controller.config

    def _load_matches_style():
        matches_style = try_load_dict_from_user_stylit()
        return matches_style

    def try_load_dict_from_user_stylit():
        stylit_path = resolve_path_stylit()
        if not os.path.exists(stylit_path):
            return None
        return load_dict_from_user_stylit(stylit_path)

    def resolve_path_stylit():
        cfg_key_fpath = 'editor.stylit_fpath'
        return config[cfg_key_fpath]

    def load_dict_from_user_stylit(stylit_path):
        matches_style = create_configobj(stylit_path, nickname='stylit')
        if matches_style is None:
            return None
        compile_eval_rules(matches_style, stylit_path)
        return matches_style

    def compile_eval_rules(matches_style, stylit_path):
        # Each section may optionally contain one code/eval component. Compile
        # it now to check for errors, with the bonus that it's cached for later
        # ((lb): not that you'd likely notice any change in performance with or
        # without the pre-compile).
        for section, rules in matches_style.items():
            if 'eval' not in rules:
                continue
            try:
                rules['__eval__'] = compile(
                    source=rules['eval'],
                    filename='<string>',
                    # Specifying 'eval' because single expression.
                    # Could use 'exec' for sequence of statements.
                    mode='eval',
                )
            except Exception as err:
                msg = _("compile() failed on 'eval' from “{0}” in “{1}”: {2}").format(
                    section, stylit_path, str(err),
                )
                dob_in_user_warning(msg)

    # ***

    return _load_matches_style()


# ***

def create_configobj(conf_path, nickname=''):
    try:
        return ConfigObj(
            conf_path,
            interpolation=False,
            write_empty_values=False,
        )
    except ConfigObjError as err:
        # Catches DuplicateError, and other errors, e.g.,
        #       Parsing failed with several errors.
        #       First error at line 55.
        msg = _("Failed to load {0} config at “{1}”: {2}").format(
            nickname, conf_path, str(err),
        )
        dob_in_user_warning(msg)
        return None




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Facts Carousel"""

from gettext import gettext as _

__all__ = (
    'render_carousel_help',
    'NUM_HELP_PAGES',
)


def render_carousel_help():
    # (lb): Is this sacrilegious, importing from that which installed *us*?
    # - I'll at least move the import inside the method. Then if this library
    #   is used outside dob (h. unlikely), the blowup will be localised.
    from dob import get_version as get_version_dob

    # FIXME/2020-04-01: Revisit this. Some commands changed; some were never implemented!
    carousel_help = _(
        """ ┏━━━━━━━━━ NAVIGATION ━━━━━━━━━┳━━━━ EDITING ━━━━┳━━━━━━━ INTERVAL ━━━━━━━━┓
 ┃ → / ←    Next/Previous Fact  ┃  [e] Edit Fact  ┃   Add/Subtract 1 min.   ┃
 ┃ j / k      Same as → / ←     ┠─────────────────╂─────────────────────────┨
 ┃ ↑ / ↓    Move Cursor Up/Down ┃    Or edit:     ┃  To Start: Shift → / ←  ┃
 ┃ h / l      Same as ↑ / ↓     ┃  [a]  act@gory  ┃  To End:    Ctrl → / ←  ┃
 ┃ PgUp     Move Cursor Up/Down ┃  [t]  tagslist  ┃  To Both:               ┃
 ┃  PgDn      by pageful        ┃  [d]  descript  ┃       Ctrl-Shift → / ←  ┃
 ┣━━━━━━━━━━━━━┳━━━━━━━━━━━┳━━━━┻━━━━━┳━━━━━━━━━━━┻━━━━━━━━━┳━━━━━━━━━━━━━━━┫
 ┃  [?] Read   ┃   Ctrl-S  ┃  Ctrl-Q  ┃  [c-p] Split Fact ½ ┃   [u]   Undo  ┃
 ┃  More Help  ┃    Save   ┃   Exit   ┃  [c-e] Empty Fact   ┃  [c-r]  Redo  ┃
 ┣━━━━━━━━━━━━━┻━━━━━━━━━━━┻━━━━┳━━━━━┻━━━━━━━━━━━┳━━━━━━━━━┻━━━━━━━━━━━━━━━┫
 ┃ [g-g]    Jump to First Fact  ┃ H A M S T E R   ┃    H A M S T E R        ┃
 ┃  [G]     Jump to Final Fact  ┃  H A M S T E R  ┃     H A M S T E R       ┃
 ┠──────────────────────────────╂─────────────────┸─────────────────────────┨
 ┃ [Home]   First line Descript ┃                                           ┃
 ┃ [End]    Bottom of Descript. ┃  dob v.{dob_vers: <34} ┃
 ┣━━━━━━━━━━━━━┳━━━━━━━━━━━━┳━━━┻━━━━━┯━━━━━━━━━┯━━━━━━━━━┳━━━━━━━━━━━━━━━━━┫
 ┃  [?] Close  ┃  [q] Easy  ┃  [c-c]  │  [c-x]  │  [c-v]  ┃   [c-z]  Undo   ┃
 ┃  this Help  ┃    Exit    ┃   Copy  │   Cut   │  Paste  ┃   [c-y]  Redo   ┃
 ┗━━━━━━━━━━━━━┻━━━━━━━━━━━━┻━━━━━━━━━┷━━━━━━━━━┷━━━━━━━━━┻━━━━━━━━━━━━━━━━━┛

        """.format(
            dob_vers=get_version_dob()[:34],
        ).rstrip()
    )
    return carousel_help


NUM_HELP_PAGES = 2




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Fact-editing “Clipboard” manager."""

from gettext import gettext as _

__all__ = (
    'ClipboardEdit',
)


class ClipboardEdit(object):
    """"""
    def __init__(self, edits_manager):
        self.controller = edits_manager.controller
        self.redo_undo = edits_manager.redo_undo
        self._clipboard = {}
        self.reset_paste()

    @property
    def clipboard(self):
        return self._clipboard

    @clipboard.setter
    def clipboard(self, clipboard):
        if self._clipboard != clipboard:
            self.paste_cnt = None
        self._clipboard = clipboard

    def reset_paste(self):
        self.paste_cnt = None

    # ***

    def copy_activity(self, curr_edit):
        self.clipboard = {'activity': curr_edit.activity, }

    def copy_tags(self, curr_edit):
        self.clipboard = {'tags': curr_edit.tags, }

    def copy_description(self, curr_edit):
        self.clipboard = {'description': curr_edit.description, }

    def copy_fact(self, curr_edit):
        self.clipboard = {'fact': curr_edit.copy(), }

    # ***

    def paste_copied_meta(self, edit_fact, reset_fact):
        """"""
        def _paste_copied_meta(edit_fact):
            if not self.clipboard:
                return None
            self.controller.affirm(len(self.clipboard) == 1)
            for paste_what, paste_val in self.clipboard.items():
                pasted_what = paste_copied_what(edit_fact, paste_what, paste_val)
            return pasted_what

        def paste_copied_what(edit_fact, paste_what, paste_val):
            if paste_what == 'activity':
                edit_fact.activity = paste_val
                pasted_what = _('activity')
            elif paste_what == 'tags':
                edit_fact.tags = paste_val
                pasted_what = _('tags')
            elif paste_what == 'description':
                edit_fact.description = paste_val
                pasted_what = _('description')
            else:
                self.controller.affirm(paste_what == 'fact')
                # MAYBE: Add a paste-all meta option? Or is cycle-pasting ok?
                #  edit_fact.activity = paste_val.activity
                #  edit_fact.tags = paste_val.tags
                #  edit_fact.description = paste_val.description
                pasted_what = paste_copied_fact(edit_fact, paste_val)
            return pasted_what

        def paste_copied_fact(edit_fact, paste_val):
            restore_fact_pre_paste(edit_fact)
            self.paste_cnt = paste_cnt_increment(edit_fact, paste_val)
            if self.paste_cnt == 0:
                edit_fact.activity = paste_val.activity
                edit_fact.tags = paste_val.tags
                pasted_what = _('activity and tags')
            elif self.paste_cnt == 1:
                edit_fact.activity = paste_val.activity
                pasted_what = _('activity')
            elif self.paste_cnt == 2:
                edit_fact.tags = paste_val.tags
                pasted_what = _('tags')
            else:
                self.controller.affirm(self.paste_cnt == 3)
                edit_fact.activity = paste_val.activity
                edit_fact.tags = paste_val.tags
                edit_fact.description = paste_val.description
                pasted_what = _('activity, tags, and description')
            return pasted_what

        def restore_fact_pre_paste(edit_fact):
            if self.paste_cnt is None:
                return
            reset_fact(edit_fact)

        def paste_cnt_increment(edit_fact, paste_val):
            paste_cnt = None
            if self.paste_cnt == 3:
                self.paste_cnt = None
            if paste_cnt is None and self.paste_cnt is None:
                if (
                    edit_fact.activity == paste_val.activity
                    and edit_fact.tags == paste_val.tags
                ):
                    self.paste_cnt = 0
                else:
                    paste_cnt = 0
            if paste_cnt is None and self.paste_cnt == 0:
                if edit_fact.activity == paste_val.activity:
                    self.paste_cnt = 1
                else:
                    paste_cnt = 1
            if paste_cnt is None and self.paste_cnt == 1:
                if edit_fact.tags == paste_val.tags:
                    self.paste_cnt = 2
                else:
                    paste_cnt = 2
            if paste_cnt is None:
                self.controller.affirm(self.paste_cnt == 2)
                paste_cnt = 3
            return paste_cnt

        return _paste_copied_meta(edit_fact)




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Facts Carousel "Matter" (Fact Description) / Content area"""

from functools import update_wrapper

from prompt_toolkit.filters import Always, Never
from prompt_toolkit.widgets import Frame, TextArea

from .exceptions import catch_action_exception
from .zone_helpful import NUM_HELP_PAGES, render_carousel_help

__all__ = (
    'ZoneContent',
)


class ZoneContent(object):
    """"""
    def __init__(self, carousel):
        self.carousel = carousel
        self.showing_help = 0
        # For your convenience, attributes to eliminate one object hop.
        self.content_lexer = self.carousel.content_lexer
        self.classes_style = self.carousel.classes_style

    # ***

    def on_reset_hide_help(self):
        was_showing = self.showing_help
        self.showing_help = 0
        return was_showing != 0

    class Decorators(object):
        @classmethod
        def reset_showing_help(cls, func):
            def wrapper(obj, event, *args, **kwargs):
                # This won't redraw because we don't call rebuild_viewable.
                # So, if, e.g., user is on first Fact and clicks Left, the
                # Help will not go away, because decrement() won't refresh.
                obj.carousel.zone_manager.zone_content.on_reset_hide_help()
                obj.carousel.zone_manager.zone_lowdown.update_status(hot_notif='')
                func(obj, event, *args, **kwargs)

            return update_wrapper(wrapper, func)

    # ***

    def standup(self):
        self.scrollable_height = self.classes_style['content-height']
        self.scrollable_width = self.classes_style['content-width']
        self.enable_wrapping = self.classes_style['content-wrap']
        self.setup_scrollable()

    # ***

    def setup_scrollable(self):
        """"""
        def _setup_scrollable():
            self.content_width = calculate_width()
            self.content = content_text_area(self.content_width)
            self.scrollable_frame = Frame(
                self.content,
                # title="Fact Description",
                style='class:content-fact',
            )

            if self.content_lexer is not None:
                self.content_lexer.content_width = self.content_width

        def calculate_width():
            avail_width = self.carousel.avail_width
            if not self.scrollable_width:
                return avail_width
            return restricted_width(avail_width)

        def restricted_width(avail_width):
            formatted = self.carousel.zone_manager.zone_lowdown.format_lowdown_text()
            len_footer = sum([len(part[1]) for part in formatted])
            content_width = max(avail_width, len_footer)
            # (lb): The width excludes the border columns, so the content
            # area will actually be two columns wider. Here we subtract
            # 1 of those columns from the width, so that the footer text
            # right-aligns one column less than the description content
            # (we could subtract 2, but then the footer is flush with the
            # border column, which I do not find as visually appealing).
            content_width -= 1
            return content_width

        # FIXME/BACKLOG/2019-01-21: Another old comment:
        #   Make `lexer` style option... or config option.

        def content_text_area(content_width):
            # Layout for displaying Fact description.
            # The Frame creates the border.
            text_area = TextArea(
                text='',
                read_only=True,
                focusable=True,  # Unnecessary; included for completeness.
                width=content_width,
                # There's also:
                #   dont_extend_width=True,
                height=self.scrollable_height,
                # Scrollbar looks wonky.
                #  And it's vertical, not horizontal,
                #   even when wrap_lines=False.
                #  scrollbar=True,
                # (lb): If user wanted to have horizontal scrolling,
                # you can set wrap_lines=False so the long lines are
                # not wrapped at width, but I cannot determine how to
                # scroll right in the TextArea. There's probably a way,
                # I'm guessing. But not a feature I care much about.
                #  wrap_lines=False,
                lexer=self.content_lexer,
            )
            return text_area

        _setup_scrollable()

    # ***

    def selectively_refresh(self):
        self.rebuild_viewable()

    # ***

    def focus_content(self, focus):
        if focus:
            self.enter_content()
        # else: nothing to do on leave.
        return True

    def enter_content(self):
        self.carousel.zone_manager.layout.focus(self.content)
        self.carousel.action_manager.wire_keys_normal()

    # *** API

    @catch_action_exception
    def cursor_up_one(self, event):
        """"""
        count = self.carousel.update_handler.apply_count_multiplier()
        self.content.buffer.cursor_up(count=count)

    @catch_action_exception
    def cursor_down_one(self, event):
        """"""
        count = self.carousel.update_handler.apply_count_multiplier()
        self.content.buffer.cursor_down(count=count)

    @catch_action_exception
    def scroll_down(self, event):
        """"""
        view_height = self.view_height()
        self.content.buffer.cursor_down(view_height)
        self.reset_cursor_left_column()

    def view_height(self):
        view_height = self.scrollable_height - 1
        if self.content.buffer.document.cursor_position_row == 0:
            # If cursor is at home posit, first page down moves cursor
            # to bottom of view. So scroll additional page, otherwise
            # user would have to press PageDown twice to see more text.
            view_height *= 2
        self.carousel.controller.client_logger.debug(
            'view_height: {}'.format(view_height)
        )
        return view_height

    @catch_action_exception
    def scroll_up(self, event):
        """"""
        self.content.buffer.cursor_up(self.scrollable_height - 1)
        self.reset_cursor_left_column()

    @catch_action_exception
    def scroll_top(self, event):
        """"""
        self.content.buffer.cursor_position = 0

    @catch_action_exception
    def scroll_bottom(self, event):
        """"""
        self.content.buffer.cursor_position = len(self.content.buffer.text)
        self.reset_cursor_left_column()

    # ***

    def reset_cursor_left_column(self):
        self.content.buffer.cursor_left(
            # PPT returns a relative distance, e.g., -7, or 0 if already there.
            # A similar command, get_cursor_left_position(), return -1 or 0.
            -self.content.buffer.document.get_start_of_line_position()
        )

    # ***

    def rebuild_viewable(self):
        content_text = self.apply_scrollable_style()
        self.content.buffer.read_only = Never()
        self.content.buffer.text = content_text
        self.content.buffer.read_only = Always()
        return self.scrollable_frame.container

    def apply_scrollable_style(self):
        if self.showing_help:
            return self.apply_scrollable_style_help()
        else:
            return self.apply_scrollable_style_fact()

    def apply_scrollable_style_help(self):
        self.scrollable_frame.container.style = 'class:content-help'
        # MAYBE: (lb): It is easy to format a PPT Frame's content?
        #   (I tried passing HTML(CAROUSEL_HELP) but, uh, nope.)
        #   (This is not too important; I thought it might be nice
        #   (polishing feature) to beautify the help (even more).)
        return render_carousel_help()

    def apply_scrollable_style_fact(self):
        self.scrollable_frame.container.style = 'class:content-fact'

        curr_edit = self.carousel.edits_manager.curr_edit

        # BACKLOG: Fact InterVal GaP Interval
        # MAYBE: Don't color after edited/added?
        #        Or probably 'unsaved-fact' style will override.
        if 'interval-gap' in curr_edit.dirty_reasons:
            self.scrollable_frame.container.style += ' class:interval-gap'

        # MAYBE: Just go through dirty_reasons and add as classes...
        if 'unsaved-fact' in curr_edit.dirty_reasons:
            self.scrollable_frame.container.style += ' class:unsaved-fact'

        # (lb): So, what do we call the stylable thing we let the user style?
        # The static style is currently 'content-fact'. The text in the content
        # area being styled is the fact description, as well as the box lines
        # around it that PPT draws. The word "description" is a lot to type,
        # e.g., "fact-description", and most users probably know what "content"
        # is, so we could call it "content-fact"... or "body" is pretty familiar,
        # though the word "body" feels like a harbinger of HTML, or at least Clue.
        # (I also considered just "content" but I like the common "fact-" prefix.)
        self.carousel.add_stylable_classes(
            self.scrollable_frame.container,
            'content-fact',
            fact=curr_edit,
        )

        content_text = curr_edit.description or ''

        content_text = self.wrap_on_whitespace_maybe(content_text)

        return content_text

    def wrap_on_whitespace_maybe(self, content_text):
        # FIXME/BACKLOG/2019-01-21: Old comment:
        #   Make KeyBinding for toggling wrapping.
        #        Ideally, 3 options: wrap, no wrap, scrollbar.
        #          For now, wrap and no-wrap,
        #            (lb): b/c I do not know if horizontal scrollbar is easily doable.
        if not self.enable_wrapping:
            return content_text
        lines_wrapped = ''
        for line in content_text.splitlines():
            line_wrapped = ''
            while len(line) > 0:
                if len(line) > self.content_width:
                    space_idx = line.rfind(' ', 0, self.content_width)
                    space_idx = self.content_width if space_idx < 1 else space_idx
                    line_piece = line[:space_idx]
                    space_idx += 1  # Do not write space at beginning of next line,
                    line = line[space_idx:]
                else:
                    line_piece = line
                    line = ''
                line_wrapped += "\n" if line_wrapped else ''
                line_wrapped += line_piece
            lines_wrapped += line_wrapped + "\n"
        return lines_wrapped

    # ***

    @catch_action_exception
    def rotate_help(self, event):
        assert NUM_HELP_PAGES == 2  # Otherwise, we need to update this logic:
        self.showing_help = (self.showing_help + 1) % (NUM_HELP_PAGES + 1)
        if self.showing_help == 1:
            self.scroll_top(event)
        elif self.showing_help == 2:
            self.scroll_bottom(event)
        else:
            self.scroll_top(event)
        self.rebuild_viewable()




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer

"""
Packaging instruction for setup tools.

Refs:

  https://setuptools.readthedocs.io/

  https://packaging.python.org/en/latest/distributing.html

  https://github.com/pypa/sampleproject
"""

from setuptools import find_packages, setup

# *** Package requirements.

requirements = [
    # The best optparser of any language, tops.
    #  https://github.com/pallets/click
    #    'click',
    # Because too busy developing our own code, our fork!
    #  https://github.com/hotoffthehamster/click
    'click-hotoffthehamster >= 7.1.1, < 8',
    # INI/config parser, even better (preserves comments and ordering).
    #  https://github.com/DiffSK/configobj
    #  https://configobj.readthedocs.io/en/latest/
    'configobj >= 5.0.6',
    # Vocabulary word pluralizer.
    #  https://github.com/ixmatus/inflector
    'Inflector >= 3.0.1, < 4',
    # Amazing prompt library.
    # - Imports as prompt_toolkit.
    # https://github.com/prompt-toolkit/python-prompt-toolkit
    'prompt-toolkit >= 3.0.5, < 4',
    # For the Carousel Fact description lexer.
    #  http://pygments.org/
    'pygments >= 2.6.1, < 3',
    # Just Another EDITOR package.
    #  https://github.com/fmoo/python-editor
    # - Imports as `editor`.
    'python-editor >= 1.0.4, < 2',
    # https://github.com/grantjenks/python-sortedcontainers/
    'sortedcontainers >= 2.1.0, < 3',

    # *** Hamster packages.

    # The heart of Hamster. (Ye olde `hamster-lib`).
    #  https://github.com/hotoffthehamster/nark
    'nark >= 3.0.5, < 4',
    # The controller, config, and common output and error tossing code.
    #  https://github.com/hotoffthehamster/dob-bright
    'dob-bright >= 1.0.8, < 2',
    # The act@gory and tag prompt interface.
    #  https://github.com/hotoffthehamster/dob-prompt
    'dob-prompt >= 1.0.5, < 2',
]

# *** Minimal setup() function -- Prefer using config where possible.

# (lb): Most settings are in setup.cfg, except identifying packages.
# (We could find-packages from within setup.cfg, but it's convoluted.)

setup(
    # Run-time dependencies installed on `pip install`. To learn more
    # about "install_requires" vs pip's requirements files, see:
    #   https://packaging.python.org/en/latest/requirements.html
    install_requires=requirements,

    # Specify which package(s) to install.
    # - Without any rules, find_packages returns, e.g.,
    #     ['dob_viewer', 'tests', 'tests.dob_viewer']
    # - With the 'exclude*' rule, this call is essentially:
    #     packages=['dob_viewer']
    packages=find_packages(exclude=['tests*']),

    # Tell setuptools to determine the version
    # from the latest SCM (git) version tag.
    #
    # Note that if the latest commit is not tagged with a version,
    # or if your working tree or index is dirty, then the version
    # from git will be appended with the commit hash that has the
    # version tag, as well as some sort of 'distance' identifier.
    # E.g., if a project has a '3.0.0a21' version tag but it's not
    # on HEAD, or if the tree or index is dirty, the version might
    # be:
    #   $ python setup.py --version
    #   3.0.0a22.dev3+g6f93d8c.d20190221
    # But if you clean up your working directory and move the tag
    # to the latest commit, you'll get the plain version, e.g.,
    #   $ python setup.py --version
    #   3.0.0a31
    # Ref:
    #   https://github.com/pypa/setuptools_scm
    setup_requires=['setuptools_scm'],
    use_scm_version=True,
)




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

# (lb): This is one of the more complicated modules in dob.
# - It almost feels like it belongs in nark, but nark does not do any
#   conflict detection nor resolution, which is what this module is all
#   about.
# - I'm also aware there might be some duplicate code between the import
#   and add facts commands, but conflict resolution, and squashing and
#   extending facts, is tricky business. Especially when the API is not
#   documented nor designed extremely robustly but grew more organically.
#   - One goal when adding tests is to document -- via tests -- all the
#     add and import use cases. Run coverage on specific use cases could
#     help a developer better understand this code.

from gettext import gettext as _

from datetime import datetime, timedelta

# MAYBE/2019-12-05: (lb): Decouple dob-viewer from click, used only for term. size.
import click_hotoffthehamster as click  # merely for get_terminal_size.

from nark.helpers import fact_time
from nark.helpers.fact_time import datetime_from_clock_after, datetime_from_clock_prior
from nark.helpers.facts_fit import antecedent_fact, subsequent_fact
from nark.helpers.parse_time import parse_clock_time, parse_relative_minutes

from dob_bright.termio import barf_and_exit, click_echo, echo_block_header

from dob_viewer.crud.fact_dressed import FactDressed

from .parse_mistakes import conflict_prefix, prepare_log_msg

__all__ = (
    'mend_fact_timey_wimey',
    'mend_facts_times',
    'must_complete_times',
    'reduce_time_hint',
    'then_extend_fact',
    'unite_and_stretch',
    # Private:
    #   'insert_forcefully',
    #   'resolve_overlapping',
)


DEFAULT_SQUASH_SEP = '\n\n--\n\n'


TIME_ERROR_ZEROTH_FACT_MUST_START = _(
    'Please specify `start` for Fact being added before time existed.'
)


# ***

def reduce_time_hint(time_hint):
    if time_hint in [
        'verify_none',
        'verify_after',
        'verify_then_none',
        'verify_then_some',
        'verify_still_none',
        'verify_still_some',
    ]:
        return 'verify_none'
    elif time_hint in [
        'verify_start',
        'verify_then_start',
        'verify_still_start',
    ]:
        return 'verify_start'
    elif time_hint in [
        'verify_both',
        'verify_end',
        'verify_then',
        'verify_still',
    ]:
        return time_hint
    assert False  # Unreachable.


# ***

def then_extend_fact(controller, new_fact):
    return FactDressed(
        activity=new_fact.activity,
        start=new_fact.end,
        end=None,
        description=None,
        tags=list(new_fact.tags),
    )


# ***

# unite_and_stretch is called on create, not transcode nor edit.
def unite_and_stretch(
    controller,
    new_fact,
    time_hint,
    conflicts,
    squash_sep=DEFAULT_SQUASH_SEP,
):
    """"""
    def _unite_and_stretch():
        if time_hint not in [
            'verify_end',  # Use prev's fact end if set; or squash facts.
            'verify_then',  # Like `at`, but back-fill gap, start is time-spec or now.
            'verify_after',  # Like `then` w/o time-spec; sets start to final's end.
            'verify_still',  # Like `then`, but copy prev fact's meta; time optional.
        ]:
            return [new_fact]
        ongoing = find_ongoing_to_stretch() if time_hint == 'verify_end' else None
        if ongoing is not None:
            return squash_or_not(ongoing)
        return extend_or_not()

    def find_ongoing_to_stretch():
        if new_fact.start:
            return None
        if new_fact.deleted:
            # If you ``to`` and ongoing fact, set_start_per_antecedent
            # (part of insert_forcefully) will squash the new_fact.
            return None
        controller.affirm(new_fact.pk is None)
        # conflicts is (edited, original) tuples.
        # One conflict is marked 'stopped' if it was ongoing Fact.
        ongoings = [con for con in conflicts if 'stopped' in con[0].dirty_reasons]
        controller.affirm(len(ongoings) <= 1)
        return ongoings[0] if ongoings else None

    def squash_or_not(ongoing):
        if time_hint == 'verify_end':
            ongoing.squash(new_fact, squash_sep)
            return [ongoing]
        else:
            # Normally, verify_then and verify_after apply to latest, stopped
            # Fact, and not an ongoing fact. In the case of the latter, don't
            # do anything; there's no time gap to process.
            return [new_fact]

    def extend_or_not():
        final_fact = controller.find_latest_fact()

        if final_fact is None:
            return [new_fact]
        new_fact_or_two = [new_fact]

        if time_hint == 'verify_end':
            controller.affirm(final_fact.end is not None)
            new_fact.start = final_fact.end
        elif time_hint in ['verify_then', 'verify_still']:
            if new_fact.start is None:
                # User did not specify then- or still-time. Use previous end.
                controller.affirm(final_fact.end is not None)
                new_fact.start = final_fact.end
            if final_fact.end is None:
                controller.affirm(new_fact.start is not None)
                final_fact.end = new_fact.start
            elif new_fact.start != final_fact.end:
                fill_fact = then_extend_fact(controller, final_fact)
                fill_fact.end = new_fact.start
                new_fact_or_two.insert(0, fill_fact)
            # else, start == end, so nothing to fill.
            if time_hint == 'verify_still':
                new_fact.activity = final_fact.activity
                new_fact.tags = list(final_fact.tags)
        elif time_hint == 'verify_after':
            new_fact.start = final_fact.end
        return new_fact_or_two

    return _unite_and_stretch()


# ***

def mend_facts_times(controller, fact, time_hint, skip_store=False):
    """"""

    def _mend_facts_times():
        # The fact is considered "temporary", or open, if the user did not
        # specify an end time, and if there's no Fact following the new Fact.
        open_start, open_end = new_fact_fill_now(fact, time_hint, controller.now)
        conflicts = []
        if not skip_store:
            conflicts = insert_forcefully(controller, fact, DEFAULT_SQUASH_SEP)

        # Note that end may be None for ongoing Fact.
        # Verify that start > end, if neither are None.
        fact_time.must_not_start_after_end((fact.start, fact.end))

        if open_start:
            fact.start = None
        if open_end:
            fact.end = None

        return conflicts

    def new_fact_fill_now(fact, time_hint, now):
        # We might temporarily set the end time to look for overlapping
        # facts, so remember if we need to leave the fact open.
        open_start = False
        open_end = False

        if time_hint == 'verify_none':
            controller.affirm(not fact.start)
            controller.affirm(not fact.end)
            fact.start = now
            open_end = True
        elif time_hint == 'verify_after':
            controller.affirm(fact.start)
            controller.affirm(not fact.end)
            open_end = True
        elif time_hint == 'verify_both':
            controller.affirm(fact.start and fact.end)
        elif time_hint == 'verify_last':
            # The end time is optional on the latest Fact.
            open_end = not fact.end
            if not fact.start:
                fact.start = now
        elif time_hint in ['verify_start', 'verify_then', 'verify_still']:
            controller.affirm(not fact.end)
            open_end = True
            if not fact.start:
                fact.start = now
        else:
            controller.affirm(time_hint == 'verify_end')
            controller.affirm(not fact.start)
            open_start = True
            if not fact.end:
                fact.end = now

        return open_start, open_end

    # ***

    return _mend_facts_times()


# ***

def must_complete_times(
    controller,
    new_facts,
    progress=None,
    leave_blanks=False,
    other_edits={},
    suppress_barf=False,
):
    """
    NOTE: new_facts must be ordered list of facts, or conflicts will happen.
      FIXME: also, write this helpdoc, eh?
    """

    # ***

    def _must_complete_times():
        if not new_facts:
            return []

        conflicts = []

        # FIXME/BACKLOG/2019-01-22: The logic here is not exactly correct,
        #   but for most current use cases, it works -- but if you were to
        #   edit an existing fact, and to specify different times for it,
        #   the call to from_other_edits_maybe() could set ante_fact or
        #   seqt_fact to the edited fact, but the edited fact is not
        #   necessarily an antecedent or a subsequent fact of the group!
        # SOLUTION: Mimic FactsManager's facts_mgr_fact_dec/_inc, and treat
        #   new_facts (a contiguous block of facts) as a time window, and
        #   create other time windows for any edited facts, and then manage
        #   all the time window groups here in this function when fixing times.
        #   E.g., if user edits a Fact and changes its time, send 1 object
        #   to this function, a sorted_contiguous_facts_list(), that contains
        #   2 groups: 1 time window blocking where the edited fact originally
        #   occupied; and 1 time window containing the newly edited fact.
        #   Then call FactsManager to get antecedent and subsequent facts
        #   (instead of calling controller.facts.subsequent/antecedent
        #      and then calling from_other_edits_maybe,
        #    like we currently do here).

        ante_fact = antecedent_fact(controller.facts, new_facts, controller.now)
        ante_fact = from_other_edits_maybe(ante_fact)

        seqt_fact = subsequent_fact(controller.facts, new_facts)
        seqt_fact = from_other_edits_maybe(seqt_fact)

        # Clean up relative clock times first (e.g., given "12:34",
        # assume date is fact's other time's date,
        # or prev time's end's date,
        # or next time's start's date).
        fix_clock_times_relative(new_facts, ante_fact, seqt_fact, conflicts)

        # Clean up relative minutes times next.
        fix_delta_times_relative(new_facts, ante_fact, seqt_fact, conflicts)

        # Finally, fill in any blanks (with adjacent times).
        if not leave_blanks:
            fix_blank_times_relative(new_facts, ante_fact, seqt_fact, conflicts)

            # Also, for import jobs (known b/c leave_blanks=False), do not
            # allow momentaneous Fact unless it has a description (because
            # a user's ``then +0`` causes transcode to back-fill Fact before
            # calling this function to resolve datetimes -- and while transcode
            # could check if the time is, e.g., ``+0``, it's more inclusive
            # to do the datetime translation and then cull these momentaneous
            # facts (so that, e.g., ``to 12:33 ...`` followed by ``then 12:33``
            # also doesn't create momentaneous).
            cull_metaless_momentaneous(new_facts, conflicts)

        # One final start < end < start ... check.
        verify_datetimes_sanity(new_facts, ante_fact, seqt_fact, conflicts)

        barf_on_overlapping_facts_new(conflicts)

        return conflicts

    # ***

    def from_other_edits_maybe(fact):
        if fact is None:
            return None
        try:
            return other_edits[fact.pk]
        except KeyError:
            return fact

    def prev_and_later(new_facts, ante_fact, seqt_fact):
        def _prev_and_later():
            prev_time = fetch_prev_fact()
            later_facts = fetch_later_facts()
            return prev_time, later_facts

        def fetch_prev_fact():
            if not ante_fact:
                return None
            prev_time = ante_fact.end or ante_fact.start
            controller.affirm(prev_time is not None)
            return prev_time

        def fetch_later_facts():
            later_facts = new_facts[0:]
            if not seqt_fact:
                return later_facts
            if not seqt_fact.start:
                controller.affirm(False)  # Caught earlier by: backend_integrity().
                raise Exception(_(
                    'Found saved fact without start time: {}'.format(seqt_fact)
                ))
            controller.affirm(isinstance(seqt_fact.start, datetime))
            later_facts += [seqt_fact]
            return later_facts

        return _prev_and_later()

    def find_next_datetime(later_facts, skip_pk=None):
        for idx, fact in enumerate(later_facts):
            if (idx == 0) and skip_pk and skip_pk == fact.pk:
                continue
            if isinstance(fact.start, datetime):
                return fact.start, fact
            if isinstance(fact.end, datetime):
                return fact.end, fact
        return None, None

    # ...

    def conflict_message(fact, msg_content):
        if suppress_barf:
            return msg_content
        return prepare_log_msg(fact, msg_content)

    # ...

    def fix_clock_times_relative(new_facts, ante_fact, seqt_fact, conflicts):
        progress and progress.click_echo_current_task(_('Fixing relative times...'))
        prev_time, later_facts = prev_and_later(new_facts, ante_fact, seqt_fact)
        for fact in new_facts:
            controller.affirm(fact is later_facts[0])
            later_facts.pop(0)
            prev_time = fix_clock_time_relative(
                fact, 'start', prev_time, later_facts, conflicts
            )
            prev_time = fix_clock_time_relative(
                fact, 'end', prev_time, later_facts, conflicts
            )

    def fix_clock_time_relative(fact, which, prev_time, later_facts, conflicts):
        dt_fact = getattr(fact, which)
        if dt_fact is not None:
            if isinstance(dt_fact, datetime):
                prev_time = dt_fact
            else:
                controller.affirm(isinstance(dt_fact, str))
                clock_time = parse_clock_time(dt_fact)
                if clock_time is not None:
                    prev_time = infer_datetime_from_clock(
                        clock_time, fact, which, prev_time, later_facts, conflicts
                    )
        return prev_time

    def infer_datetime_from_clock(
        clock_time, fact, which, prev_time, later_facts, conflicts,
    ):
        dt_suss = None

        at_oppo = 'end' if which == 'start' else 'start'
        dt_oppo = getattr(fact, at_oppo)
        if isinstance(dt_oppo, datetime):
            # Prefer relative to fact's other time.
            if which == 'start':
                dt_suss, err = infer_from_clock_prior(fact, dt_oppo, clock_time)
            else:
                dt_suss, err = infer_from_clock_after(fact, dt_oppo, clock_time)
        elif prev_time is not None:
            # If fact's other time not set, prefer earlier clock time.
            dt_suss, err = infer_from_clock_after(fact, prev_time, clock_time)
        elif not later_facts:
            # First fact ever! being entered. Use now, now.
            #   2019-01-11 17:43: On add first new fact. At least from:
            #     py.test --pdb -vv -k test_start_add_new_fact tests/
            # Note that we do not use infer_from_clock_after, which would add a day
            # to avoid a conflict, e.g., do not do this:
            #   dt_suss, err = infer_from_clock_after(fact, controller.now, clock_time)
            dt_suss, err = infer_from_clock_prior(fact, controller.now, clock_time)
        else:
            # Last resort: go hunting for the next actual, factual real datetime.
            # FIXME/2018-05-21 12:17: (lb): Will this ever happen?
            #   Probably only if antecedent_fact not found??
            next_time, _next_fact = find_next_datetime(later_facts)
            if next_time is not None:
                dt_suss, err = infer_from_clock_prior(fact, next_time, clock_time)

        if dt_suss is not None:
            controller.affirm(err is None)
            setattr(fact, which, dt_suss)
            prev_time = dt_suss
        else:
            msg_content = _(
                'Could not decipher clock time “{}” for {}'
            ).format(clock_time, which)
            conflict_msg = conflict_message(fact, msg_content)
            conflicts.append((fact, None, conflict_msg, ))
        return prev_time

    def infer_from_clock_prior(fact, dt_oppo, clock_time):
        dt_suss = None
        err_msg = None
        try:
            dt_suss = datetime_from_clock_prior(dt_oppo, clock_time)
        except Exception as err:
            err_msg = str(err)
        return dt_suss, err_msg

    def infer_from_clock_after(fact, dt_oppo, clock_time):
        dt_suss = None
        err_msg = None
        try:
            dt_suss = datetime_from_clock_after(dt_oppo, clock_time)
        except Exception as err:
            err_msg = str(err)
        return dt_suss, err_msg

    # ...

    def fix_delta_times_relative(new_facts, ante_fact, seqt_fact, conflicts):
        progress and progress.click_echo_current_task(_('Fixing relative deltas...'))
        prev_time, later_facts = prev_and_later(new_facts, ante_fact, seqt_fact)
        for fact in new_facts:
            controller.affirm(fact is later_facts[0])
            later_facts.pop(0)
            prev_time = fix_delta_time_relative(
                fact, 'start', prev_time, later_facts, conflicts
            )
            prev_time = fix_delta_time_relative(
                fact, 'end', prev_time, later_facts, conflicts
            )

    def fix_delta_time_relative(fact, which, prev_time, later_facts, conflicts):
        dt_fact = getattr(fact, which)
        if dt_fact is not None:
            if isinstance(dt_fact, datetime):
                prev_time = dt_fact
            else:
                controller.affirm(isinstance(dt_fact, str))
                delta_mins, delta_minus = parse_relative_minutes(dt_fact)
                if delta_mins is not None:
                    prev_time = infer_datetime_from_delta(
                        delta_mins,
                        delta_minus,
                        fact,
                        which,
                        prev_time,
                        later_facts,
                        conflicts,
                    )
        return prev_time

    def infer_datetime_from_delta(
        delta_mins, delta_minus, fact, which, prev_time, later_facts, conflicts,
    ):
        # If -delta, relative to *next* time; if +delta, relative to *prev* time.
        dt_suss = None
        if delta_minus is False:
            if prev_time is not None:
                dt_suss = prev_time + timedelta(minutes=delta_mins)
            # else, we'll add a conflict below.
        elif which == 'start' and isinstance(fact.end, datetime):
            # NOTE: delta_mins < 0, so add the delta (to subtract it).
            dt_suss = fact.end + timedelta(minutes=delta_mins)
        else:
            next_time, _next_fact = find_next_datetime(later_facts)
            if next_time is not None:
                # NOTE: delta_mins is negative, so add to next_time.
                dt_suss = next_time + timedelta(minutes=delta_mins)

        if dt_suss is not None:
            setattr(fact, which, dt_suss)
            prev_time = dt_suss
        else:
            msg_content = _(
                'Could not interpret delta time “{}” for {}'
            ).format(delta_mins, which)
            conflict_msg = conflict_message(fact, msg_content)
            conflicts.append((fact, None, conflict_msg, ))
        return prev_time

    # ...

    def fix_blank_times_relative(new_facts, ante_fact, seqt_fact, conflicts):
        progress and progress.click_echo_current_task(_('Fixing blank times...'))
        prev_time, later_facts = prev_and_later(new_facts, ante_fact, seqt_fact)
        for idx, fact in enumerate(new_facts):
            controller.affirm(fact is later_facts[0])
            later_facts.pop(0)
            controller.affirm(fact.start or fact.end)
            first_fact = idx == 0
            prev_time = fix_blank_time_relative(
                fact, 'start', prev_time, later_facts, conflicts, first_fact,
            )
            prev_time = fix_blank_time_relative(
                fact, 'end', prev_time, later_facts, conflicts, first_fact,
            )

    def fix_blank_time_relative(
        fact, which, prev_time, later_facts, conflicts, first_fact,
    ):
        dt_fact = getattr(fact, which)
        if isinstance(dt_fact, datetime):
            prev_time = dt_fact
        elif dt_fact:
            # If not None and not a datetime, still a string,
            # meaning was unable to resolve earlier.
            conflicts_add_unresolved_relative(fact, which, conflicts, dt_fact)
        else:
            dt_suss = suss_datetime(which, prev_time, later_facts)
            if dt_suss is not None:
                setattr(fact, which, dt_suss)
                prev_time = dt_suss
            else:
                conflicts_add_unsussed_relative(fact, which, conflicts, first_fact)
        return prev_time

    def suss_datetime(which, prev_time, later_facts):
        dt_suss = None
        if which == 'start':
            dt_suss = prev_time
        elif which == 'end':
            dt_suss, _next_fact = find_next_datetime(later_facts)
            if dt_suss is None:
                dt_suss = controller.store.now
        return dt_suss

    def conflicts_add_unresolved_relative(fact, which, conflicts, dt_fact):
        msg_content = _(
            'Could not translate relative time “{}” for {}'
        ).format(dt_fact, which)
        conflict_msg = conflict_message(fact, msg_content)
        conflicts.append((fact, None, conflict_msg, ))

    def conflicts_add_unsussed_relative(fact, which, conflicts, first_fact):
        if which == 'start' and first_fact:
            # Caller will (should) call mend_facts_times later...
            return
        msg_content = _(
            'Could not infer time left blank for {}'
        ).format(which)
        conflict_msg = conflict_message(fact, msg_content)
        conflicts.append((fact, None, conflict_msg, ))

    # ...

    def cull_metaless_momentaneous(new_facts, conflicts):
        allow_momentaneous = controller.config['time.allow_momentaneous']
        new_culled = []
        for idx, fact in enumerate(new_facts):
            if (
                (fact.start == fact.end)
                and ((not fact.description) or (not allow_momentaneous))
            ):
                if fact.description:
                    msg_content = _(
                        "Momentaneous Fact found, but allow_momentaneous=False"
                    )
                    conflict_msg = conflict_message(fact, msg_content)
                    conflicts.append((fact, None, conflict_msg, ))
                continue
            new_culled.append(fact)
        new_facts[:] = new_culled

    # ...

    def verify_datetimes_sanity(new_facts, ante_fact, seqt_fact, conflicts):
        """"""
        progress and progress.click_echo_current_task(_('Verifying sanity times...'))
        prev_time, later_facts = prev_and_later(new_facts, ante_fact, seqt_fact)
        prev_fact = ante_fact
        for idx, fact in enumerate(new_facts):
            controller.affirm(fact is later_facts[0])
            later_facts.pop(0)
            n_datetimes = 0

            if not fact.start:
                # Rather than adding it again, e.g.,
                #   conflicts.append((
                #     fact, None, _('Could not determine start of new fact'))
                #   )
                # just verify we already caught it.
                is_edge_fact = idx == 0
                affirm_datetimes_missing_already_caught(fact, conflicts, is_edge_fact)
            elif isinstance(fact.start, datetime):
                if prev_time and fact.start < prev_time:
                    msg_content = _('New fact starts before previous fact ends')
                    conflict_msg = conflict_message(fact, msg_content)
                    conflicts.append((fact, prev_fact, conflict_msg, ))
                prev_time = fact.start
                n_datetimes += 1
            # else, a string that was unparsed; and conflict already added.

            if not fact.end:
                is_edge_fact = idx == (len(new_facts) - 1)
                affirm_datetimes_missing_already_caught(fact, conflicts, is_edge_fact)
            elif isinstance(fact.end, datetime):
                next_time, next_fact = find_next_datetime(later_facts, fact.pk)
                if next_time and fact.end > next_time:
                    msg_content = _('New fact ends after next fact starts')
                    conflict_msg = conflict_message(fact, msg_content)
                    conflicts.append((fact, next_fact, conflict_msg, ))
                prev_time = fact.end
                n_datetimes += 1
            # else, a string that was unparsed; and conflict already added.

            if n_datetimes == 2 and fact.start > fact.end:
                msg_content = _('New fact starts after it ends/ends before it starts')
                conflict_msg = conflict_message(fact, msg_content)
                conflicts.append((fact, None, conflict_msg, ))

            prev_fact = fact

    def affirm_datetimes_missing_already_caught(fact, conflicts, is_edge_fact):
        if leave_blanks:
            return
        found_it = list(filter(lambda conflict: fact is conflict[0], conflicts))
        controller.affirm(len(found_it) > 0 or is_edge_fact)

    # ...

    def barf_on_overlapping_facts_new(conflicts):
        if (not conflicts) or (suppress_barf):
            return

        colorful = controller.config['term.use_color']

        for fact, other, reason in conflicts:
            # other might be None, another new Fact, or ante_fact or seqt_fact.
            if other and other.pk:
                prefix = 'Saved'
                other_pk = '#{}: '.format(other.pk)
            else:
                prefix = 'New'
                other_pk = ''
            echo_block_header(
                _('{} Fact Datetime Conflict!'.format(prefix)),
                full_width=True,
            )
            click_echo()
            click_echo(fact.friendly_diff(fact, truncate=True))
            click_echo()
            click_echo(reason)
            if other:
                # FIXME/2018-06-12: (lb): Subtract edges; this is too much.
                cut_width = click.get_terminal_size()[0]

                click_echo('\n{}: {}{}'.format(
                    conflict_prefix(_('Compare')),
                    other_pk,
                    other.friendly_str(colorful=colorful, cut_width=cut_width),
                ))
            click_echo()
        msg = _(
            'Could not determine, or did not validate, start and/or stop'
            ' for one or more new Facts.'
            '\nScroll up for details.'
        )
        crude = len(new_facts) > 1
        barf_and_exit(msg, crude=crude)

    return _must_complete_times()


# ***

# FIXME/TESTING/2019-01-22: Test adding fact when there's an ongoing Fact.
#   - Test new Fact with and without start.
#     - A new Fact without start gets squashed into current ongoing Fact?
#     - A new Fact with a start ends the current ongoing.
#   - Three ways to add new Fact: 1.) single CLI command; 2.) Carousel; 3.) import.
#   - Or does other code handle ongoing Fact, I don't remember...

# NOTE/2019-01-22: insert_forcefully checks one Fact being edited against
# the store only, and it ignores any other edited Facts you might have.
# The current use cases define this restriction:
# - The command line only lets the user edit one Fact at a time;
# - The Carousel manages it own set of edited Facts and does not
#   need (call) this function (and it will not let user change a
#   Fact's start or end time if the new Fact's times would fully
#   shadow another Fact's times);
# - The import function will let a user edit multiple Facts, but
#   it die-complains if *any* Fact conflicts, so its use of this
#   function is only in the context of a single Fact vs. the db.

def insert_forcefully(controller, fact, squash_sep=''):
    """
    Insert the possibly open-ended Fact into the set of logical
    (chronological) Facts, possibly changing the time frames of,
    or removing, other Facts.

    Args:
        fact (nark.Fact):
            The Fact to insert, with either or both ``start`` and ``end`` set.

    Returns:
        list: List of edited ``Facts``, ordered by ``start``.

    Raises:
        ValueError: If start or end time is not specified and cannot be
            deduced by other Facts in the system.
    """
    allow_momentaneous = controller.config['time.allow_momentaneous']

    def _insert_forcefully(facts, fact):
        # Steps:
        #   Find fact overlapping start.
        #   Find fact overlapping end.
        #   Find facts wholly contained between start and end.
        #   Return unique set of facts indicating edits and deletions.

        conflicts = []
        conflicts += find_conflict_at_edge(facts, fact, 'start')
        conflicts += find_conflict_at_edge(facts, fact, 'end')
        conflicts += find_conflicts_during(facts, fact)

        edited_conflicts = resolve_overlapping(
            fact, conflicts, squash_sep, allow_momentaneous,
        )

        return edited_conflicts

    # ***

    def find_conflict_at_edge(facts, fact, ref_time):
        conflicts = []
        find_edge = False
        fact_time = getattr(fact, ref_time)
        if fact_time:  # fact.start or fact.end
            conflicts = facts.surrounding(fact_time)
            if conflicts:
                if len(conflicts) != 1:
                    controller.client_logger.warning(_(
                        "Found more than one Fact ({} total) at: '{}'"
                        .format(len(conflicts), fact_time))
                    )
            else:
                find_edge = True
        else:
            find_edge = True
        if find_edge:
            assert not conflicts
            conflicts = inspect_time_boundary(facts, fact, ref_time)
        return conflicts

    def inspect_time_boundary(facts, fact, ref_time):
        conflict = None
        if ref_time == 'start':
            if fact.start is None:
                # MAYBE/SIMPLIFY/2019-01-22: Lots of comments in insert_forcefully
                #   about perhaps being redundant, and using just must_complete_times.
                # Here's one such old comment:
                #   # FIXME: rely on must_complete_times.
                #   # we could probably assert that fact.start is not None,
                #   #  and remove set_start_per_antecedent

                # (lb): This code branch is a squash fact situation!
                # - If you start an active, uncompleted Fact, and then try,
                #   e.g., `dob to 1000`, or `dob to +1h`, you'll get here.
                # - Another easy here: `dob after`.
                # - 2019-12-06: I changed set_start_per_antecedent to not fail if
                #   the new Fact has no start, so that it'll get squashed with the
                #   active fact later. (Just noting now in case breaks soon after.)
                conflict = set_start_per_antecedent(facts, fact)
            else:
                conflict = facts.starting_at(fact)
        else:
            assert ref_time == 'end'
            if fact.end is None:
                # This use case will get you here, because no end time:
                #   echo "2020-01-28 15:38: hello" | dob import
                # EXPLAIN/2020-01-28: But fact still saved as ongoing fact...
                #   interest. I think this is Desired Behavior, but curious
                #   what removes the end time (which set_end_per_subsequent
                #   sets to now).
                # - 2020-01-28: Old comment might still have some relevance:
                # FIXME/TESTME/2019-01-22: Is this path possible? Will code
                # have called must_complete_times first, and this situation
                # have been remedied? See comment above, before function def,
                # that talks about the different use cases: Codify all this
                # blather in tests! (I'm curious how the functions here in
                # insert_forcefully, set_end_per_subsequent, as well as
                # set_start_per_antecedent, compare to must_complete_times,
                # which has its own start and end setting code.)
                set_end_per_subsequent(facts, fact)
            else:
                # LATER/BACKLOG/LOWLOWPRIORITY/2019-01-22: Momentaneous:
                #   If supporting momentaneous, could have multiple Facts
                #   with same ending_at (and not just multiple versions
                #   of the same Fact (with all but 1 deleted).
                conflict = facts.ending_at(fact)
        conflicts = [conflict] if conflict else []
        return conflicts

    def set_start_per_antecedent(facts, fact):
        assert fact.start is None

        # Find a Fact with start < fact.end.
        ref_fact = facts.antecedent(fact)

        # MAYBE/SIMPLIFY/2019-01-22: Merge this code into must_complete_times?
        # - See "redundant" comment below, too.
        # - 6 mo. old comment says:
        #   - FIXME: Make this error a conflict in must_complete_times,
        #             then remove set_start_per_antecedent.

        if not ref_fact:
            raise ValueError(TIME_ERROR_ZEROTH_FACT_MUST_START)

        # Because we called surrounding and got nothing, we know that
        # found_fact.end < fact.end; or that found_fact.end is None,
        # a/k/a, the ongoing Fact.
        conflict = None
        if ref_fact.end is not None:
            assert ref_fact.end < fact.end
            fact.start = ref_fact.end
        else:
            # There's an ongoing Fact, and the new Fact has no start, which
            # indicates that these two facts should be squashed. (We'll create
            # an intermediate conflict now, and we'll squash the Facts later,
            # so that we include the ongoing Fact in the list of edited Facts
            # we return later.)
            assert ref_fact.start < fact.end
            conflict = ref_fact

        return conflict

    # MAYBE/SIMPLIFY/2019-01-22: Is this function redundant? Can it be removed?
    #   A comment from 6 mos. ago suggests that
    #     must_complete_times acts similar to set_end_per_subsequent,
    #     or that set_end_per_subsequent can be merged into must_complete_times.
    #   I might also not care anymore. Maybe consult coverage for the truth?

    def set_end_per_subsequent(facts, fact):
        assert fact.end is None
        ref_fact = facts.subsequent(fact)
        if ref_fact:
            assert ref_fact.start > fact.start
            fact.end = ref_fact.start
        else:
            # This is ongoing fact/current.
            controller.client_logger.debug(
                _("No end specified for Fact; assuming now.")
            )
            fact.end = controller.now  # Same as: facts.store.now
            # NOTE: for dob-on, we'll start start, then end will be
            #       a few micros later... but the caller knows to unset
            #       this Fact's end later (see: leave_open).
            #       (lb): I wrote this code and I can't quite remember
            #       why we fact to do this. I think so comparing against
            #       other Facts works....

    # ***

    def find_conflicts_during(facts, fact):
        conflicts = []
        if fact.start and fact.end:
            found_facts = facts.strictly_during(fact.start, fact.end)
            conflicts += found_facts
        return conflicts

    # ***

    return _insert_forcefully(controller.facts, fact)


# ***

def resolve_overlapping(fact, conflicts, squash_sep='', allow_momentaneous=False):
    """"""
    def _resolve_overlapping(fact, conflicts):
        seen = set()
        resolved = []
        for conflict in conflicts:
            assert conflict.pk > 0
            if fact.pk == conflict.pk:
                # Editing existing Fact may find itself in db.
                continue
            if conflict.pk in seen:
                continue
            seen.add(conflict.pk)
            original = conflict.copy()
            edited_conflicts = resolve_fact_conflict(fact, conflict)
            for edited in edited_conflicts:
                resolved.append((edited, original,))
        return resolved

    def resolve_fact_conflict(fact, conflict):
        # If the conflict is contained within another Fact, that
        # other Fact will be split in twain, so we may end up
        # with more conflicts.
        resolved = []
        if fact.start is None and conflict.end is None:
            resolve_fact_squash_fact(fact, conflict, resolved)
        elif fact.start <= conflict.start:
            resolve_fact_starts_before(fact, conflict, resolved)
        elif conflict.end is None or fact.end >= conflict.end:
            resolve_fact_ends_after(fact, conflict, resolved)
        else:
            # The new fact is contained *within* the conflict!
            resolve_fact_is_inside(fact, conflict, resolved)
        return cull_duplicates(resolved)

    def resolve_fact_squash_fact(fact, conflict, resolved):
        conflict.squash(fact, squash_sep)
        resolved.append(conflict)

    def resolve_fact_starts_before(fact, conflict, resolved):
        if fact.end <= conflict.start:
            # Disparate facts.
            return
        elif conflict.end and fact.end >= conflict.end:
            if (
                allow_momentaneous
                and (fact.start == conflict.start)
                and (conflict.start == conflict.end)
            ):
                # (lb): 0-length Fact is not surrounded by new Fact.
                #   As they say in Futurama, I'm going to allow this.
                return
            conflict.deleted = True
            conflict.dirty_reasons.add('deleted-starts_before')
        else:
            # This is either the last Fact in the database, which is still
            # open (if conflict.end is None); or fact ends before conflict
            # ends. And in either case, fact ends after conflict starts,
            # so move conflict's start to no longer conflict.
            assert conflict.start < fact.end
            conflict.start = fact.end
            conflict.dirty_reasons.add('start')
        resolved.append(conflict)

    def resolve_fact_ends_after(fact, conflict, resolved):
        if conflict.end is not None and fact.start >= conflict.end:
            # Disparate facts.
            return
        elif fact.start <= conflict.start:
            if (
                allow_momentaneous
                and (fact.end == conflict.end)
                and (conflict.start == conflict.end)
            ):
                # 0-length Fact is not surrounded by new Fact; I'll allow it.
                return
            conflict.deleted = True
            conflict.dirty_reasons.add('deleted-ends_after')
        else:
            # (lb): Here's where we might stop an ongoing fact
            # when adding a new fact.
            assert conflict.end is None or conflict.end > fact.start
            # A little hack: signal the caller if this is/was ongoing fact.
            if conflict.end is None:
                conflict.dirty_reasons.add('stopped')
            conflict.end = fact.start
            conflict.dirty_reasons.add('end')
        resolved.append(conflict)

    def resolve_fact_is_inside(fact, conflict, resolved):
        resolve_fact_split_prior(fact, conflict, resolved)
        resolve_fact_split_after(fact, conflict, resolved)

    def resolve_fact_split_prior(fact, conflict, resolved):
        # Make a copy of the conflict, to not affect resolve_fact_split_after.
        lconflict = conflict.copy()
        lconflict.split_from = conflict.pk
        # Leave lconflict.pk set so the old fact is marked deleted.
        lconflict.end = fact.start
        lconflict.dirty_reasons.add('lsplit')
        resolved.append(lconflict)

    def resolve_fact_split_after(fact, conflict, resolved):
        rconflict = conflict.copy()
        rconflict.split_from = conflict.pk
        rconflict.pk = None
        rconflict.start = fact.end
        rconflict.dirty_reasons.add('rsplit')
        resolved.append(rconflict)

    def cull_duplicates(resolved):
        seen = set()
        culled = []
        for conflict in resolved:
            if conflict in seen:
                continue
            seen.add(conflict)
            culled.append(conflict)
        return culled

    # ***

    return _resolve_overlapping(fact, conflicts)


# ***

def mend_fact_timey_wimey(controller, fact, time_hint, other_edits={}):
    """"""
    def _mend_fact_timey_wimey():
        must_complete_time(controller, fact, other_edits)
        # Fill in the start and, or, end times, maybe.
        # Possibly correct the times of 2 other Facts!
        # Or die if too many Facts are abound tonight.
        conflicts = mend_facts_times(controller, fact, time_hint)
        # Resolve conflicts from store with other edited facts being saved.
        conflicts = rebuild_conflicts(fact, conflicts, other_edits)
        new_fact_or_two = unite_and_stretch_fact_per_conflicts(conflicts)
        return new_fact_or_two, conflicts

    def unite_and_stretch_fact_per_conflicts(conflicts):
        # On `to` and `then`, combine fact and latest.
        # Note that insert_forcefully will handle ``to`` for ongoing fact;
        # otherwise unite_and_stretch handles it for ended latest.
        if fact.deleted:
            return []
        return unite_and_stretch(controller, fact, time_hint, conflicts)

    def must_complete_time(controller, fact, other_edits):
        reset_end = fact.end is None
        if fact.end is None:
            fact.end = controller.now
        new_facts = [fact, ]
        must_complete_times(
            controller,
            new_facts,
            leave_blanks=True,
            other_edits=other_edits,
        )
        assert len(new_facts) == 1
        fact.end = None if reset_end else fact.end

    def rebuild_conflicts(fact, conflicts, other_edits):
        if not other_edits:
            return conflicts
        culled_conflicts = []
        for conflict in conflicts:
            verify_conflict(fact, conflict, other_edits, culled_conflicts)
        return culled_conflicts

    def verify_conflict(fact, conflict, other_edits, culled_conflicts):
        auto_edit, orig_fact = conflict
        assert auto_edit.pk == orig_fact.pk
        try:
            edit_fact = other_edits[orig_fact.pk]
        except KeyError:
            # Fact was not separately edited, so conflict from store stands.
            culled_conflicts.append(conflict)
        else:
            new_conflict = recheck_conflict(edit_fact, orig_fact)
            if new_conflict is not None:
                culled_conflicts.append(new_conflict)

    def recheck_conflict(edit_fact, orig_fact):
        # Both Facts should have a start, and only 1 may be unended.
        assert fact.start and edit_fact.start
        assert fact.end or edit_fact.end
        if edit_fact.end is None:
            # If edit_fact is ongoing, should start after fact.
            if edit_fact.start >= fact.end:
                return None
        elif fact.end is None:
            # If fact is ongoing, should start after edit_fact.
            if fact.start >= edit_fact.end:
                return None
        elif edit_fact.start == edit_fact.end:
            # Momentaneous can only happen on boundary.
            if (edit_fact.start <= fact.start) or (edit_fact.start >= fact.end):
                return None
        elif fact.start == fact.end:
            # Momentaneous can only happen on boundary.
            if (fact.start <= edit_fact.start) or (fact.start >= edit_fact.end):
                return None
        elif (edit_fact.end <= fact.start) or (edit_fact.start >= fact.end):
            # Both facts complete, neither is momentaneous, and ranges are distinct.
            return None
        # If we made it here, whoa! Unexpected time conflict after using carousel?
        new_conflict = (edit_fact, orig_fact, )
        return new_conflict

    # ***

    return _mend_fact_timey_wimey()




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Fact Editing State Machine"""

from .clipboard_edit import ClipboardEdit
from .facts_manager import FactsManager
from .group_chained import sorted_facts_list
from .redo_undo_edit import RedoUndoEdit
from .start_end_edit import StartEndEdit

__all__ = (
    'EditsManager',
)


class EditsManager(object):
    """"""
    def __init__(
        self,
        controller,
        edit_facts=None,
        orig_facts=None,
        dirty_callback=None,
        error_callback=None,
    ):
        self.controller = controller
        self.setup_editing(edit_facts, orig_facts)
        self._dirty_callback = dirty_callback
        self.error_callback = error_callback

    # ***

    def setup_editing(self, edit_facts, orig_facts):
        """"""
        self.setup_container(edit_facts, orig_facts)
        self.setup_edit_facts(edit_facts)
        self.setup_review_confirmation()
        self.setup_edit_help()

    # ***

    def setup_container(self, edit_facts, orig_facts):
        def _setup_container():
            orig_lkup = orig_facts_lookup(orig_facts)
            apply_orig_facts(edit_facts, orig_lkup)
            self.conjoined = FactsManager(
                self.controller,
                on_jumped_fact=self.jumped_fact,
                on_insert_fact=self.insert_fact,
            )
            self.add_facts(edit_facts)

        def orig_facts_lookup(orig_facts):
            orig_lkup = orig_facts or {}
            if isinstance(orig_lkup, list):
                orig_lkup = {fact.pk: fact for fact in orig_lkup}
            return orig_lkup

        def apply_orig_facts(edit_facts, orig_lkup):
            for edit_fact in edit_facts:
                apply_orig_fact(edit_fact, orig_lkup)

        def apply_orig_fact(edit_fact, orig_lkup):
            self.controller.affirm(edit_fact.orig_fact is None)
            try:
                # NOTE: For facts from the store loaded on start, this is the
                # only reference to them. (For facts loaded later from the
                # store, they're stored on self.edit_facts only until they're
                # edited, then they, too, become just an orig_fact reference
                # on an edit_fact, or a fact on the undo/redo stack.
                edit_fact.orig_fact = orig_lkup[edit_fact.pk]
            except KeyError:
                edit_fact.orig_fact = 0
            if edit_fact.orig_fact:
                # FIXME/2019-01-20: TEST NEW PATH
                self.controller.affirm(False)  # (lb): Not been here yet.
                # FIXME/2019-01-22: If you make empty group, should make time-gap,
                # so group has at least 1 Fact! Should also manage gap-time entry as
                # time-space changes (for any group? or will the inc/dec code figure
                # that out otherwise?? too many edge cases!).
                assign_orig_fact(edit_fact, orig_lkup)
                # Note that claim_time_span changes the group key, but
                # the group has not been added to the conjoined.groups
                # container yet (so no need to use self.fact_group_rekeyed).
                self.conjoined.claim_time_span(*edit_fact.orig_fact.times)

        def assign_orig_fact(edit_fact, orig_lkup):
            self.controller.affirm(edit_fact.orig_fact is not edit_fact)
            self.controller.affirm(edit_fact.orig_fact.orig_fact is None)
            edit_fact.orig_fact.orig_fact = 0

        _setup_container()

    def add_facts(self, more_facts):
        for fact in more_facts:
            if fact.orig_fact is None:
                # Rather than be self-referential and set, say, fact.orig_fact = fact,
                # we use a magic placeholder, 0, that happens to be non-truthy, and
                # indicates that this fact is the original, unedited copy of itself.
                fact.orig_fact = 0  # As opposed to None, means original, unedited copy.
        self.conjoined.add_facts(more_facts)

    def setup_edit_facts(self, edit_facts):
        # Dirty facts, on stand up, will only include import facts, or fact
        # entered on command line; but will ignore fact read from store, e.g.,
        # `dob edit -1` will start up with an empty self.edit_facts (and the
        # one fact loaded from the store will be held in the conjoined.groups).
        self.edit_facts = {fact.pk: fact for fact in edit_facts if fact.dirty}

    # ***

    def setup_review_confirmation(self):
        self.verify_fact_pks = set([fact.pk for fact in self.conjoined.facts])
        self.viewed_fact_pks = set()

    # ***

    def setup_edit_help(self):
        self.setup_redo_undo()
        self.setup_clipboard()
        self.setup_time_edit()

    def setup_redo_undo(self):
        self.redo_undo = RedoUndoEdit(self)

    def setup_clipboard(self):
        self.clipboard = ClipboardEdit(self)

    def setup_time_edit(self):
        self.time_edit = StartEndEdit(self)

    # ***

    def dirty_callback(self):
        if self._dirty_callback is None:
            return
        self._dirty_callback(self)

    @property
    def is_dirty(self):
        return len(self.edit_facts) > 0

    # ***

    @property
    def curr_fact(self):
        return self.conjoined.curr_fact

    @curr_fact.setter
    def curr_fact(self, curr_fact):
        """"""
        self.controller.client_logger.debug(
            '\n- curr: {}'.format(curr_fact.short),
        )
        if self.conjoined.curr_fact is not curr_fact:
            self.clipboard.reset_paste()
        self.conjoined.curr_fact = curr_fact
        self.viewed_fact_pks.add(curr_fact.pk)

    def insert_fact(self, gap_fact):
        self.redo_undo.update_undo_altered([gap_fact], append=True)

    def jumped_fact(self, jump_fact):
        # Jump to shim to the setter.
        self.curr_fact = jump_fact

    @property
    def user_viewed_all_new_facts(self):
        return self.verify_fact_pks.issubset(self.viewed_fact_pks)

    @property
    def curr_edit(self):
        """
        Returns the currently edited fact, or the original fact if
        nothing being edited. Because this might return the original,
        uneditable fact, he caller is not expected to edit the fact.
        (See editable_fact() for retrieving the editable equivalent
        of this function.)
        """
        try:
            return self.edit_facts[self.curr_fact.pk]
        except KeyError:
            return self.curr_fact

    @property
    def curr_orig(self):
        return self.curr_fact.orig_fact or self.curr_fact

    # ***

    @property
    def edit_fact_count(self):
        return len(self.edit_facts)

    @property
    def edit_fact_index(self):
        return sorted_facts_list(self.edit_facts.values()).index(self.curr_fact)

    # ***

    @property
    def prepared_facts(self):
        """
        Returns list of edited and new facts to persist (to database, export file, etc.).
        """
        prepared_facts_from_edit = sorted_facts_list(self.edit_facts.values())
        prepared_facts_from_view = [
            fact for fact in self.conjoined.facts if fact.dirty
        ]
        self.controller.affirm(prepared_facts_from_edit == prepared_facts_from_view)
        return prepared_facts_from_edit

    # ***

    def editable_fact(self):
        # Copy Fact on demand for user to edit, if we haven't made one already.
        try:
            edit_fact = self.edit_facts[self.curr_fact.pk]
            # On dob-import, the original import facts are put in self.edit_facts.
            # So make a copy if what's in edit_facts is the original. (Later, when
            # update_lookups is called via update_edited_fact to update edit_facts,
            # we'll reference this new copy (and this new copy will keep the orig_fact
            # object alive).)
            if not edit_fact.orig_fact:
                self.controller.affirm(edit_fact.orig_fact == 0)
                edit_fact = edit_fact.copy()
            elif edit_fact is self.curr_fact:
                # (lb): The FactsManager fact-groups are wired with the latest
                # edit of the Fact, so if the user wants to edit a Fact again, be
                # sure to pass them a Fact copy, lest they edit the one that is
                # already part of a fact-group. (Editing a Fact once it is part of
                # a SortedKeyList causes a number of problems. For one, the group
                # key, conjoined.groups[n].sorty_times, returns a different value,
                # but, internally, the _maxes value is unchanged. This causes the
                # group to behave oddly. And the only way to ensure that _maxes is
                # updated is to remove and reinsert the Fact, but to remove the
                # Fact, we need to be able to identify it. Hence, we cannot have
                # other code inadvertently mucking with a Fact once it's been
                # added to the group-fact container. So return a copy.)
                edit_fact = edit_fact.copy()
        except KeyError:
            # Use the latest version of the fact, not orig_fact.
            edit_fact = self.curr_fact.copy()
            self.controller.affirm(
                (edit_fact.orig_fact is self.curr_fact)
                or (edit_fact.orig_fact is self.curr_fact.orig_fact)
            )
            # (lb): Fact might later be placed in self.edit_facts via
            # update_edited_fact if the operation that needs edit_fact
            # actually changes it.
        return edit_fact

    def undoable_editable_fact(self, what, edit_fact=None):
        # Always push the Fact onto the undo stack, should it be edited.
        # The caller calls update_redo_undo_and_conjoined() after editing,
        # which might pop the Fact if the user did not edit anything.
        if edit_fact is None:
            edit_fact = self.editable_fact()
        was_fact = edit_fact.copy()
        self.add_undoable([was_fact], what)
        # Caller is responsible for calling this class' apply_edits later.
        return edit_fact

    def apply_edits(self, *edit_facts):
        # Called on edit-time, and after carousel prompts user for edits.
        # edit_facts is a list of 1 or 2 Facts: the edited Fact, and then
        # maybe a prev or next Fact, if one exists and was affected by a
        # change to the edited Fact's start or end time.
        edit_facts = list(filter(None, edit_facts))
        applied_edits = self.update_redo_undo_and_conjoined(edit_facts)
        if not applied_edits:
            self.controller.affirm(edit_facts[0] == self.curr_fact)
            return
        self.dirty_callback()
        # Show the first edited Fact (and ensure Carousel showing a wired
        # Fact, in case what was curr_fact was removed during the edit).
        self.curr_fact = self.conjoined.locate_wired(edit_facts[0])

    def add_undoable(self, was_facts, what):
        self.redo_undo.add_undoable(was_facts, what)

    def update_redo_undo_and_conjoined(self, edit_facts):
        # edit_facts is 1 or 2 items: current fact, maybe followed by neighbor.
        last_edits = self.redo_undo.remove_undo_if_nothing_changed(edit_facts)
        if last_edits is None:
            return False

        self.manage_edited_dirty_flags(edit_facts)

        self.conjoined.apply_edits(edit_facts, last_edits)

        # Update undo with edited Facts.
        self.redo_undo.update_undo_altered(edit_facts)

        return True

    def manage_edited_dirty_flags(self, edit_facts):
        for idx, edit_fact in enumerate(edit_facts):
            # Only unmark interval-gap if this is the first fact in edit_facts,
            # which indicates user deliberately edited gap-fact; as opposed
            # to if gap-fact is later in edit_facts, and then it was edited
            # by time adjust of another fact. Feels like could be better way.
            is_oldest_fact = idx == 0
            self.manage_edited_dirty_deleted(edit_fact, undelete=is_oldest_fact)
            self.manage_edited_edit_facts(edit_fact)

    def manage_edited_dirty_deleted(self, edit_fact, undelete=False):
        edit_fact.dirty_reasons.add('unsaved-fact')
        if not undelete:
            return

        # If user edited gap-fact, ensure returned by prepared_facts;
        # and clear its highlight.
        edit_fact.dirty_reasons.discard('interval-gap')

        # An interval-gap is not necessarily marked deleted,
        #  but if edited, we should make sure it no longer is.
        # Also, if any Fact was marked deleted, but user edited
        #  it, should also make sure not deleted.
        edit_fact.deleted = False

    def manage_edited_edit_facts(self, edit_fact):
        orig_fact = edit_fact.orig_fact or edit_fact
        self.update_edited_fact(edit_fact, orig_fact)

    def update_edited_fact(self, edit_fact, orig_fact):
        self.controller.affirm(edit_fact is not orig_fact)
        self.controller.affirm((orig_fact == 0) or (edit_fact.pk == orig_fact.pk))
        if edit_fact.dirty:
            # Update or add reference to latest edit.
            self.edit_facts[edit_fact.pk] = edit_fact
        elif orig_fact != 0:
            self.controller.affirm(not orig_fact.dirty)
            try:
                # Forget edited fact that's no longer different than orig.
                self.edit_facts.pop(orig_fact.pk)  # Ignoring: popped fact.
            except KeyError:
                pass

    # ***

    def stand_up(self):
        def _standup():
            ensure_backed_up()
            assert len(self.conjoined.groups) > 0
            assert len(self.conjoined.groups[0]) > 0
            ensure_view_facts()
            self.conjoined.place_time_rifts()
            self.curr_fact = self.conjoined.find_first_dirty()

        def ensure_backed_up():
            # Create a just-in-case backup file to capture unsaved edits. We'll
            # call periodically during editor session, and can start session by
            # calling it, too (stand_up is called at start of carousel.gallop).
            self.dirty_callback()

        def ensure_view_facts():
            """
            Ensure at least 1 fact is loaded, because
            there is no empty Carousel state!.
            """
            if len(list(self.conjoined.facts)):
                return
            at_least_load_latest_fact()

        def at_least_load_latest_fact():
            self.controller.affirm(len(self.conjoined.groups) == 0)
            latest_fact = self.controller.facts.antecedent(
                ref_time=self.controller.now,
            )
            self.controller.affirm(latest_fact is not None)
            latest_fact.orig_fact = 0
            # FIXME: When latest_fact is None => what's empty carousel state?
            self.add_facts([latest_fact])

        _standup()

    # ***

    def toss_last_edit(self):
        undone = self.redo_undo.undo_last_edit(restore_facts=None)
        return undone

    def undo_last_edit(self):
        undone = self.redo_undo.undo_last_edit(self.restore_facts)
        return undone

    def redo_last_undo(self):
        redone = self.redo_undo.redo_last_undo(self.restore_facts)
        return redone

    def restore_facts(self, pristine, altered):
        # 2019-01-31: (lb): I think we can skip manage_edited_dirty_deleted
        # (managing 'interval-gap') because that's encoded in redo-undo Facts.

        for edit_fact in pristine:
            self.controller.affirm(edit_fact.orig_fact.pk != 0)
            self.update_edited_fact(edit_fact, edit_fact.orig_fact)

        self.conjoined.apply_edits(edit_facts=pristine, last_edits=altered)
        # Jump to the "main" Fact that was edited.
        self.curr_fact = self.conjoined.locate_wired(pristine[0])
        self.dirty_callback()

    # ***

    def fact_copy_activity(self):
        self.clipboard.copy_activity(self.curr_edit)

    def fact_copy_tags(self):
        self.clipboard.copy_tags(self.curr_edit)

    def fact_copy_description(self):
        self.clipboard.copy_description(self.curr_edit)

    def fact_copy_fact(self):
        self.clipboard.copy_fact(self.curr_edit)

    # ***

    def paste_copied_meta(self):
        """"""
        if not self.clipboard.clipboard:
            return None
        edit_fact = self.undoable_editable_fact(what='paste-copied')
        pasted_what = self.clipboard.paste_copied_meta(
            edit_fact, reset_fact=self.reset_copied_meta,
        )
        self.apply_edits(edit_fact)

        return pasted_what

    def reset_copied_meta(self, edit_fact):
        # The clipboard has a cycle mechanism (self.clipboard.paste_cnt) that
        # cycles over different attributes to paste as the user repeats the
        # same command. E.g., first press pastes activity. Next press pastes
        # the tags. Third pastes description. And fourth pastes everything.
        # Between each paste, the fact is reset, so only one attribute gets
        # changed, and there's only one undo item created.

        # The first undo is the one we created in paste_copied_meta.
        _latest_changes = self.redo_undo.undo.pop()
        self.controller.affirm(len(_latest_changes.pristine) == 1)
        self.controller.affirm(_latest_changes.pristine[0] == edit_fact)

        # The second undo is the one created the last time this method called.
        before_paste = self.redo_undo.undo.pop()
        self.controller.affirm(len(before_paste.pristine) == 1)

        # Reset edit_fact in place.
        restore_fact = before_paste.pristine[0]
        edit_fact.activity = restore_fact.activity
        edit_fact.tags = restore_fact.tags
        edit_fact.description = restore_fact.description
        self.controller.affirm(edit_fact.orig_fact)

        # Start a new undo (sets UndoRedoTuple.pristine with copy of edit_fact).
        self.redo_undo.add_undoable([edit_fact.copy()], before_paste.what)
        # EditManager.paste_copied_meta calls its apply_edits after
        # calling this method, which ensures UndoRedoTuple.altered is set.

    # ***

    def edit_time_adjust(self, *args, **kwargs):
        self.time_edit.edit_time_adjust(*args, **kwargs)

    # ***

    def editable_fact_prev(self, edit_fact):
        prev_fact = self.jump_fact_dec()
        if prev_fact is None:
            return None
        edit_prev = self.editable_fact()
        _curr_fact = self.jump_fact_inc()
        self.controller.affirm(_curr_fact.pk == edit_fact.pk)
        return edit_prev

    def editable_fact_next(self, edit_fact):
        next_fact = self.jump_fact_inc()
        if next_fact is None:
            return None
        edit_next = self.editable_fact()
        _curr_fact = self.jump_fact_dec()
        self.controller.affirm(_curr_fact.pk == edit_fact.pk)
        return edit_next

    # ***

    def jump_fact_dec(self, count=1):
        """"""
        for idx in range(count):
            prev_fact = self.conjoined.jump_fact_dec()
            if prev_fact is None:
                return prev_fact
            elif prev_fact.dirty:
                self.update_edited_fact(prev_fact, prev_fact.orig_fact)
        return prev_fact

    def jump_fact_inc(self, count=1):
        """"""
        for idx in range(count):
            next_fact = self.conjoined.jump_fact_inc()
            if next_fact is None:
                return next_fact
            elif next_fact.dirty:
                self.update_edited_fact(next_fact, next_fact.orig_fact)
        return next_fact

    # ***

    def jump_day_dec(self, days=1):
        """"""
        return self.conjoined.jump_day_dec(days=days)

    def jump_day_inc(self, days=1):
        """"""
        return self.conjoined.jump_day_inc(days=days)

    # ***

    def jump_rift_dec(self):
        """"""
        self.conjoined.jump_rift_dec()

    def jump_rift_inc(self):
        """"""
        self.conjoined.jump_rift_inc()

    # ***

    def save_edited_facts(self):
        """"""
        # 2019-01-23 22:28: (lb): I wrote this quick in the past hour.
        # Seems to work. Guess we'll see how stable it is!

        def _save_edited_facts():
            curr_fact = self.curr_fact
            # 2019-01-23 20:46: Just assume the Carousel handled conflicts?
            # LATER/BACKLOG: What about if store changed in background?
            #   - Or would changed Facts have PK marked deleted?
            #     Would error propagate on changed db?
            edited_facts = self.prepared_facts
            ignore_pks = [fact.pk for fact in edited_facts]
            keep_fact, saved_facts = save_edited_trustworthy(
                edited_facts, ignore_pks,
            )
            keep_fact = reset_editing(keep_fact, saved_facts, curr_fact)
            # Return fact for zone_manager to jump to.
            return keep_fact, saved_facts

        def save_edited_trustworthy(edited_facts, ignore_pks):
            keep_fact = None
            saved_facts = []
            for edit_fact in edited_facts:
                new_fact = save_edited_fact(edit_fact, ignore_pks)
                if new_fact is None:
                    return save_edited_fact_failed()
                saved_facts.append(new_fact)
                if edit_fact is self.curr_fact:
                    keep_fact = new_fact
                affirm_saved_edited_fact(edit_fact, new_fact)
            return keep_fact, saved_facts

        def save_edited_fact_failed():
            # Something went wrong, and we displayed an error.
            # Return now, and leave the store in a Bad State.
            # Users are encouraged to keep their data stores
            # under revision control so that they can recover
            # from blunders such as these. And we should make
            # sure our code is tough and resilient and stable.
            # I.e., we don't need to bother handling this error
            # better; just don't cause an error.
            return None, None  # Short-circuit return!

        def save_edited_fact(edit_fact, ignore_pks):
            # (lb): SIMILAR: edits_manager.save_edited_fact, create.save_fact.
            if edit_fact.pk and edit_fact.pk < 0:
                edit_fact.pk = None
            if edit_fact.pk is None and edit_fact.deleted:
                self.controller.client_logger.debug(
                    'Deleted fact: {}'.format(edit_fact.short)
                )
                return []
            try:
                return self.controller.facts.save(
                    edit_fact, ignore_pks=ignore_pks,
                )
            except Exception as err:
                import traceback
                self.controller.client_logger.debug(traceback.format_exc())
                # A failure on a CLI command (without PPT interface) might do:
                #   import traceback
                #   traceback.print_exc()
                #   dob_in_user_exit(str(err))
                # But Carousel has a popup message handler.
                self.error_callback(errmsg='Failed to save fact!\n\n  “{}”'.format(err))
                return None

        def affirm_saved_edited_fact(edit_fact, new_fact):
            # (lb): It's easier to reset editing than to try to update state.
            #   So just a few affirmations, and then moving along.
            #   (The caller will return the saved curr_fact, and we'll
            #   rebuild the Carousel with that one Fact. Everything else
            #   will be rebuilt from scratch.)
            if edit_fact.pk:
                # PK is different for saved fact, and old fact is marked deleted;
                #   except for active (an ongoing) Fact, which retains its ID.
                if edit_fact.pk != new_fact.pk:
                    self.controller.affirm(new_fact.pk >= edit_fact.pk)
                    self.controller.affirm(edit_fact.deleted)
                else:
                    self.controller.affirm(not edit_fact.deleted)
                self.controller.affirm(self.edit_facts[edit_fact.pk] is edit_fact)
            else:
                # PK is None, so new Fact.
                self.controller.affirm(not edit_fact.deleted)
                self.controller.affirm(not new_fact.deleted)
                self.controller.affirm(new_fact.pk > 0)
            self.controller.affirm(new_fact.orig_fact is None)

        def reset_editing(keep_fact, saved_facts, curr_fact):
            if not saved_facts:
                return curr_fact
            if keep_fact is None:
                keep_fact = curr_fact
                curr_fact.orig_fact = None
                curr_fact.next_fact = None
                curr_fact.prev_fact = None
            # Because saving creates new Fact IDs, and because
            # it'd be a pain to update either all affected
            # Facts' IDs (think not only prepared_facts/conjoined.facts,
            # but also redo_undo.undo, redo_undo.redo, conjoined.by_pk,
            # and most importantly, all the Facts' next_fact and prev_fact
            # pointers!), reset every place that has a reference to any
            # Facts.
            # See also: self.setup_edit_help()
            keep_fact.orig_fact = None
            keep_fact.next_fact = None
            keep_fact.prev_fact = None
            self.setup_editing(edit_facts=[keep_fact], orig_facts=[])
            self.curr_fact = keep_fact
            return keep_fact

        return _save_edited_facts()




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

import re

__all__ = (
    'namilize',
)

# SYNC_ME: Use similar re as PPT. From prompt_toolkit/styles/style.py::
#  CLASS_NAMES_RE = re.compile(r'^[a-z0-9.\s_-]*$')  # This one can't contain a comma!
# - We use the regex for intra-line character substitution, so drop the ^, *, and $.
# - (lb): Also, drop the \s, because we're forming a single classname.
#   - I think technically classnames may contain spaces, but if we don't replace
#     spaces, PTK parsing raises (at ValueError("Wrong color format %r" % text)).
CLASS_NAME_NOTSET = re.compile(r'[^a-z0-9._-]')

# MAYBE/2020-03-31: Behave like, say, PyPI.org, and condense multiple dashes into one?
# - Would this make writing the styling config simpler? Or more restrictive?
# - Without a good use case, I'll note the idea here, but that's it. -lb.
#
#  COMPRESS_DASHES = re.compile(r'-+')


def namilize(name):
    """Normalizes a string so it can be used as a prompt_toolkit (PPT) classname.
    """
    classname = CLASS_NAME_NOTSET.sub('-', name.lower())
    # See comment above: To condense dashes, or not. For now, not!
    #  classname = COMPRESS_DASHES.sub('-', classname)
    return classname




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

""""""

from functools import update_wrapper

__all__ = (
    'catch_action_exception',
)


def catch_action_exception(func):
    def wrapper(self, event=None, *args, **kwargs):
        try:
            return func(self, event=event, *args, **kwargs)
        except Exception as err:  # noqa: F841
            # F841 local variable '...' is assigned to but never used

            if not self.carousel.controller.config['dev.catch_errors']:
                # MAYBE/2019-01-21: Display warning and silently recover.
                #  But really, harden the code, and do not expect this path.
                raise

            import traceback
            traceback.print_stack()
            traceback.print_exc()
            # We're gonna die anyway, so undo terminal raw mode, if PPT did so.
            import os
            os.system('stty sane')
            import pdb
            pdb.set_trace()
            raise

    return update_wrapper(wrapper, func)

